import typing

from collections import Counter

from dbnd._core.constants import (
    DescribeFormat,
    RunState,
    SystemTaskName,
    TaskRunState,
    UpdateSource,
)
from dbnd._core.current import is_verbose
from dbnd._core.run.run_ctrl import RunCtrl
from dbnd._core.settings.core import CoreConfig
from dbnd._core.tracking.schemas.tracking_info_objects import TaskRunEnvInfo
from dbnd._core.utils.basics.text_banner import TextBanner


if typing.TYPE_CHECKING:
    from dbnd._core.run.databand_run import DatabandRun


class DescribeRun(RunCtrl):
    @property
    def tracker(self):
        return self.run.tracker

    def _add_tasks_info(self, b):
        run = self.run
        reused = sum(
            tr.is_reused and not tr.is_skipped_as_not_required for tr in run.task_runs
        )
        optimizations = []
        if reused:
            optimizations.append(
                "There are {completed} reused tasks.".format(completed=reused)
            )
        task_skipped_as_not_required = sum(
            tr.is_reused and tr.is_skipped_as_not_required for tr in run.task_runs
        )
        if task_skipped_as_not_required:
            optimizations.append(
                " {skipped} tasks are not required by any uncompleted task "
                "that is essential for your root task.".format(
                    skipped=task_skipped_as_not_required
                )
            )
        show_more = is_verbose()
        task_runs = run.task_runs
        if not show_more:  # show only non system
            task_runs = [tr for tr in run.task_runs if not tr.task.task_is_system]
            if not task_runs:
                task_runs = run.task_runs

        states = Counter(tr.task_run_state for tr in task_runs if tr.task_run_state)

        tasks = [("total", len(task_runs))]
        tasks += [(k.value, v) for k, v in states.items()]

        b.column_properties("TASKS", tasks)
        if optimizations:
            b.column("RUN OPTIMIZATION", " ".join(optimizations))
        return b

    def run_banner(self, msg, color="white", show_run_info=False, show_tasks_info=True):
        b = TextBanner(msg, color)
        run = self.run  # type: DatabandRun
        ctx = run.context
        task_run_env = ctx.task_run_env  # type: TaskRunEnvInfo
        driver_task = run.driver_task_run.task

        b.column("TRACKER URL", run.run_url, skip_if_empty=True)
        b.column("TRACKERS", CoreConfig().tracker)
        if show_tasks_info and run.is_orchestration and driver_task.is_driver:
            self._add_tasks_info(b)

        if run.root_run_info.root_run_uid != run.run_uid:
            b.column(
                "ROOT TRACKER URL", run.root_run_info.root_run_url, skip_if_empty=True
            )
            b.column("ROOT UID URL", run.root_run_info.root_run_uid, skip_if_empty=True)

        if run.scheduled_run_info:
            b.column_properties(
                "SCHEDULED",
                [
                    ("scheduled_job", run.scheduled_run_info.scheduled_job_uid),
                    ("scheduled_date", run.scheduled_run_info.scheduled_date),
                    ("dag_run_id", run.scheduled_run_info.scheduled_job_dag_run_id),
                ],
            )

        if show_run_info:
            b.new_line()
            run_params = [
                ("user", task_run_env.user),
                ("run_uid", "%s" % run.run_uid),
                ("env", run.env.name),
            ]
            b.column("RUN", b.f_simple_dict(run_params))
            b.column(
                "LOG",
                b.f_simple_dict(
                    [
                        ("local", driver_task.local_driver_log),
                        ("remote", driver_task.remote_driver_root),
                    ]
                ),
            )
            b.column("USER CODE VERSION", task_run_env.user_code_version)
            b.column("CMD", task_run_env.cmd_line)

            if run.is_orchestration:
                b.column(
                    "EXECUTE",
                    b.f_simple_dict(
                        [
                            ("TASK_EXECUTOR", run.task_executor_type),
                            ("PARALLEL", run.parallel),
                            ("SUBMIT_DRIVER", run.submit_driver),
                            ("SUBMIT_TASKS", run.submit_tasks),
                        ],
                        skip_if_empty=True,
                    ),
                    skip_if_empty=True,
                )
            if task_run_env.user_data and task_run_env.user_data != "None":
                b.column("USER DATA", task_run_env.user_data, skip_if_empty=True)
            b.new_line()

        failed_task_runs = [
            task_run
            for task_run in run.task_runs
            if task_run.task_run_state in TaskRunState.direct_fail_states()
        ]
        if failed_task_runs:
            f_msg = "\n\t".join(tr.task.task_id for tr in failed_task_runs)
            b.column("FAILED", f_msg)

        if run.root_task_run and run.is_orchestration:
            b.column("TASK_BAND", run.root_task_run.task.task_band)

        b.new_line()

        return b.getvalue()

    def run_banner_for_finished(self):
        if self.run._run_state == RunState.SUCCESS:
            root_task_run = self.run.root_task_run
            root_task = root_task_run.task
            msg = "Your run has been successfully executed!"
            if self.run.duration:
                msg = (
                    "Your run has been successfully executed in %s" % self.run.duration
                )
            return "\n%s\n%s\n" % (
                root_task.ctrl.banner(
                    "Main task '%s' is ready!" % root_task.task_name,
                    color="green",
                    task_run=root_task_run,
                ),
                self.run_banner(msg, color="green",),
            )

        return self.run_banner("Your run has failed!", color="red", show_run_info=True)

    def run_banner_for_submitted(self):
        msg = (
            "Your run has been submitted! Please check submitted job in UI\n"
            " * Use '--interactive' to have blocking run (wait for driver completion)\n"
            " * Use '--local-driver' (env.submit_driver=False) to run your driver locally\n"
        )
        return self.run_banner(
            msg, color="green", show_run_info=False, show_tasks_info=False
        )

    def get_error_banner(self):
        # type: ()->str
        err_banners = []
        run = self.run
        err_banners.append(self.run_banner_for_finished())

        failed_task_runs = []
        for task_run in run.task_runs:
            if (
                task_run.last_error
                or task_run.task_run_state in TaskRunState.direct_fail_states()
            ):
                failed_task_runs.append(task_run)

        if len(failed_task_runs) > 1:
            # clear out driver task, we don't want to print it twice
            failed_task_runs = [
                tr
                for tr in failed_task_runs
                if tr.task.task_name not in SystemTaskName.driver_and_submitter
            ]

        for task_run in failed_task_runs:
            if task_run.task_run_state == TaskRunState.CANCELLED:
                msg_header = "Task has been terminated!"
            else:
                msg_header = "Task has failed!"
            msg = task_run.task.ctrl.banner(
                msg=msg_header, color="red", task_run=task_run
            )
            err_banners.append(msg)

        err_banners.append(
            self.run_banner(
                "Your run has failed! See more info above.",
                color="red",
                show_run_info=False,
            )
        )
        return u"\n".join(err_banners)


def print_tasks_tree(root_task, task_runs, describe_format=DescribeFormat.short):
    from dbnd._core.task_ctrl.task_dag_describe import DescribeDagCtrl

    completed = {tr.task.task_id: tr.is_reused for tr in task_runs}
    run_describe_dag = DescribeDagCtrl(
        root_task, describe_format, complete_status=completed
    )
    run_describe_dag.tree_view(describe_format=describe_format)



from typing import Any, Callable, List, Type, TypeVar, Union, overload

import numpy
import pandas as pd

from dbnd._core.configuration.config_path import ConfigPath
from dbnd._core.constants import OutputMode
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    ParameterScope,
    _ParameterKind,
)
from dbnd._core.parameter.validators import Validator
from targets import DataTarget, Target
from targets.target_config import FileFormat, TargetConfig
from targets.types import DataList
from targets.values import ValueType

_T = TypeVar("_T")
_C = TypeVar("_C")

class ParameterFactory(object):
    parameter: ParameterDefinition
    def __init__(self, parameter: ParameterDefinition = None): ...
    @overload
    def value(
        self,
        default: _T = ...,
        description: str = ...,
        significant: bool = ...,
        target_config: TargetConfig = ...,
        config_path: ConfigPath = ...,
        value_type: ValueType = ...,
        system: bool = ...,
        scope: ParameterScope = ...,
        kind: _ParameterKind = ...,
        output_name: str = ...,
        output_ext: str = ...,
        name: str = ...,
        output_factory: Callable[[Any, ParameterDefinition], Any] = ...,
        flag: bool = ...,
        output_mode: OutputMode = ...,
        validator: Validator = ...,
        choices: List = ...,
        parameter_cls: Type = ...,
        load_on_build: bool = ...,
        empty_default: bool = ...,
        log_histograms: bool = ...,
        log_size: bool = ...,
        log_stats: bool = ...,
        log_schema: bool = ...,
        log_preview: bool = ...,
        log_preview_size: int = ...,
    ) -> _T: ...
    @overload
    def value(
        self,
        default: None = ...,
        significant: bool = ...,
        description: str = ...,
        scope: ParameterScope = ...,
    ) -> ParameterFactory: ...
    def __call__(
        self,
        default: Any = ...,
        name: str = ...,
        description: str = ...,
        significant: bool = ...,
        config_path: ConfigPath = ...,
        value_type: ValueType = ...,
        system: bool = ...,
        scope: ParameterScope = ...,
        kind: _ParameterKind = ...,
        target_config: TargetConfig = ...,
        output_name: str = ...,
        output_ext: str = ...,
        output_factory: Callable = ...,
        output_mode: OutputMode = ...,
        validator: Validator = ...,
        choices: List = ...,
        parameter_cls: Type = ...,
        load_on_build: bool = ...,
        empty_default: bool = ...,
        log_histograms: bool = ...,
        log_size: bool = ...,
        log_stats: bool = ...,
        log_schema: bool = ...,
        log_preview: bool = ...,
        log_preview_size: int = ...,
    ) -> ParameterFactory: ...
    def __getitem__(self, type_: Type[_C]) -> _C: ...
    def modify(self, **kwargs: Any) -> ParameterFactory: ...
    @property
    def none(self) -> ParameterFactory: ...
    ########
    # build parameter
    @property
    def _p(self) -> ParameterDefinition: ...
    def build_parameter(self, str) -> ParameterDefinition: ...
    def _build_parameter(self, str) -> ParameterDefinition: ...
    def _build_value_type(self, str) -> ValueType: ...
    ########
    # formats
    def target_config(self, target_config: TargetConfig) -> ParameterFactory: ...
    @property
    def _target_config(self) -> TargetConfig: ...
    @property
    def no_ext(self) -> DataTarget: ...
    @property
    def gz(self) -> ParameterFactory: ...
    @property
    def gzip(self) -> ParameterFactory: ...
    @property
    def parquet(self) -> ParameterFactory: ...
    @property
    def json(self) -> ParameterFactory: ...
    @property
    def hdf5(self) -> ParameterFactory: ...
    @property
    def txt(self) -> ParameterFactory: ...
    @property
    def csv(self) -> ParameterFactory: ...
    @property
    def pickle(self) -> ParameterFactory: ...
    @property
    def feather(self) -> ParameterFactory: ...
    @property
    def excel(self) -> ParameterFactory: ...
    @property
    def file_numpy(self) -> ParameterFactory: ...
    @property
    def require_local_access(self) -> ParameterFactory: ...
    @property
    def no_extension(self) -> ParameterFactory: ...
    @property
    def no_env_interpolation(self) -> ParameterFactory: ...
    @property
    def disable_jinja_templating(self) -> ParameterFactory: ...
    def load_options(
        self, file_format: FileFormat, **options: Any
    ) -> ParameterFactory: ...
    def save_options(
        self, file_format: FileFormat, **options: Any
    ) -> ParameterFactory: ...
    @overload
    def custom_target(self, target_factory: Type[_C], folder=False) -> _C: ...
    @overload
    def custom_target(self, target_factory: Callable, folder=False) -> Target: ...
    @overload
    def target_factory(self, target_factory: Callable, folder=False) -> Target: ...
    @property
    def folder(self) -> ParameterFactory: ...
    @property
    def folder_data(self) -> ParameterFactory: ...
    def with_flag(self, flag: Union[bool, str, None] = True) -> ParameterFactory: ...
    @property
    def prod_immutable(self) -> ParameterFactory: ...
    # kind changes
    @property
    def output(self) -> ParameterFactory: ...
    def cast(self, type_: Type[_C]) -> _C: ...
    def sub_type(self, sub_type) -> ParameterFactory: ...
    def type(self, type_) -> ParameterFactory: ...
    def parameter_cls(self, parameter_cls: ParameterDefinition) -> ParameterFactory: ...
    @property
    def data(self) -> DataTarget: ...
    def help(self, help_message: str) -> ParameterFactory: ...
    @property
    def non_significant(self) -> ParameterFactory: ...
    def config(self, section: str, name: str) -> ParameterFactory: ...
    def description(self, help_message: str) -> ParameterFactory: ...
    def default(self, value: _T) -> _T: ...
    def enum(self, enum: _T) -> _T: ...
    def choices(self, choices) -> ParameterFactory: ...
    def numerical(
        self, min_value: Union[int, float], max_value: Union[int, float]
    ) -> ParameterFactory: ...
    def custom_datetime(self, interval=1, start=None) -> ParameterFactory: ...
    def nested_config(self, config_cls: _C) -> _C: ...
    @property
    def c(self) -> ParameterFactory: ...
    @property
    def system(self) -> ParameterFactory: ...
    @property
    def target(self) -> DataTarget: ...
    @property
    def path_str(self) -> str: ...
    @property
    def pandas_dataframe(self) -> pd.DataFrame: ...
    @property
    def data_list_str(self) -> DataList[str]: ...
    @property
    def data_list(self) -> DataList: ...
    @property
    def numpy_array(self) -> numpy.ndarray: ...
    @property
    def overwrite(self) -> ParameterFactory: ...

@overload
def value(
    default: None = ...,
    significant: bool = ...,
    description: str = ...,
    scope: ParameterScope = ...,
) -> ParameterFactory: ...
@overload
def value(
    default: _T,
    significant: bool = ...,
    description: str = ...,
    scope: ParameterScope = ...,
) -> _T: ...

PARAMETER_FACTORY: ParameterFactory
data: ParameterFactory
output: ParameterFactory
parameter: ParameterFactory

def get_parameter_cls(value_type: ValueType) -> ParameterDefinition: ...
def build_parameter(
    parameter: Union[ParameterFactory, ParameterDefinition], context: str
) -> ParameterDefinition: ...
def parameter_config(p: ParameterDefinition, key: str) -> ParameterFactory: ...



from dbnd._core.current import get_databand_context
from dbnd._core.tracking.backends.channels.abstract_channel import TrackingChannel
from dbnd._core.tracking.backends.channels.marshmallow_mixin import MarshmallowMixin
from dbnd._core.utils.basics.memoized import cached


class TrackingWebChannel(MarshmallowMixin, TrackingChannel):
    """Json API client implementation."""

    @property
    @cached()
    def client(self):
        return get_databand_context().databand_api_client

    def _handle(self, name, data):
        return self.client.api_request("tracking/%s" % name, data)

    def is_ready(self):
        return self.client.is_ready()

    def __str__(self):
        return "Web"



from dbnd import parameter, project_path
from dbnd._core.task.config import Config


class FatWheelConfig(Config):
    _conf__task_family = "bdist_zip"
    _conf__help_description = "Configuration for building fat wheel file that contains project and third party requirements"
    package_dir = parameter.c(
        default=project_path(),
        description="Full path to the directory of the package that contains setup.py",
    )[str]
    requirements_file = parameter.c(
        description="Full path to the requirements file (if any)", default=None
    )[str]



import logging
import os
import signal

from dbnd._core.utils.timezone import utcnow


logger = logging.getLogger(__name__)


def safe_signal(signalnum, handler):
    # wraps signal cmd, so it doesn't throws an exception
    # while running on windows/multithreaded environment
    try:
        return signal.signal(signalnum, handler)
    except Exception:
        logger.info("Failed to set an alert handler for '%s' signal", signalnum)
    return None


SIGQUIT_DUMP_DIR = "/tmp"


def sigquit_handler__dump_stack(sig, frame):
    """Helps debug deadlocks by printing stacktraces when this gets a SIGQUIT
    e.g. kill -s QUIT <PID> or CTRL+\
    """
    dump_trace(dump_file=True)
    return True


def dump_trace(dump_file=None):
    import threading, traceback, sys, os

    try:

        def _p(msg):
            sys.__stderr__.write("%s\n" % msg)

        _p("Dumping stack traces for all threads in PID {}".format(os.getpid()))
        id_to_name = dict([(th.ident, th.name) for th in threading.enumerate()])
        code = []
        for thread_id, stack in sys._current_frames().items():
            code.append(
                "\n# Thread: {}({})".format(id_to_name.get(thread_id, ""), thread_id)
            )
            for filename, line_number, name, line in traceback.extract_stack(stack):
                code.append(
                    'File: "{}", line {}, in {}'.format(filename, line_number, name)
                )
                if line:
                    code.append("  {}".format(line.strip()))
        traceback_data = "\n".join(code)
        _p("%s\n" % traceback_data)
        if dump_file is True:
            dump_file = os.path.join(
                SIGQUIT_DUMP_DIR,
                "dbnd.dump.%s.txt" % (utcnow().strftime("%Y%m%d-%H%M%S")),
            )
        if dump_file:
            with open(dump_file, "wt") as df_fp:
                df_fp.write(traceback_data)
            _p("Stack has been dumped into {}".format(dump_file))
        return traceback_data
    except Exception as e:
        print(
            "Couldn't report stack traces after reciving SIGQUIT! Exception: %s", str(e)
        )


def register_sigquit_stack_dump_handler():
    logger.warning(
        "SIGQUIT signal registered! Use `kill -%s %s` or CTRL+\\ for stacktrace dump."
        % (signal.SIGQUIT, os.getpid())
    )
    signal.signal(signal.SIGQUIT, sigquit_handler__dump_stack)



# -*- coding: utf-8 -*-


from dbnd._core.cli.main import (
    dbnd_cmd,
    dbnd_run_cmd,
    dbnd_run_cmd_main,
    main as dbnd_main,
)
from dbnd._core.commands import log_dataframe, log_duration, log_metric, log_metrics
from dbnd._core.configuration.config_path import ConfigPath
from dbnd._core.configuration.config_readers import override
from dbnd._core.configuration.config_store import ConfigMergeSettings
from dbnd._core.configuration.dbnd_config import config, config_deco
from dbnd._core.context.bootstrap import dbnd_bootstrap
from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.current import (
    current_task,
    current_task_run,
    dbnd_context,
    get_databand_context,
    get_databand_run,
)
from dbnd._core.decorator.dbnd_decorator import (
    band,
    data_source_pipeline,
    pipeline,
    task,
)
from dbnd._core.failures import dbnd_handle_errors
from dbnd._core.inplace_run.inplace_run_manager import dbnd_run_start, dbnd_run_stop
from dbnd._core.parameter.parameter_builder import data, output, parameter
from dbnd._core.parameter.parameter_definition import (
    ParameterDefinition,
    ParameterScope,
)
from dbnd._core.plugin.dbnd_plugins import hookimpl
from dbnd._core.task.config import Config
from dbnd._core.task.data_source_task import DataSourceTask
from dbnd._core.task.pipeline_task import PipelineTask
from dbnd._core.task.python_task import PythonTask
from dbnd._core.task.task import Task
from dbnd._core.task_build import task_namespace
from dbnd._core.task_build.task_context import current
from dbnd._core.task_build.task_namespace import auto_namespace, namespace
from dbnd._core.task_build.task_registry import register_config_cls, register_task
from dbnd._core.task_ctrl.task_relations import as_task
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.tracking.python_tracking import (
    track_functions,
    track_module_functions,
    track_modules,
)
from dbnd._core.utils.project.project_fs import (
    databand_lib_path,
    databand_system_path,
    project_path,
    relative_path,
)
from dbnd.tasks import basics
from targets import _set_patches


from dbnd._core.configuration.environ_config import (  # isort:skip
    get_dbnd_project_config,
)


get_dbnd_project_config().validate_init()  # isort:skip


dbnd_config = config
__all__ = [
    "hookimpl",
    # context management
    "new_dbnd_context",
    "current",
    "dbnd_context",
    "current_task",
    "current_task_run",
    "get_databand_run",
    "get_databand_context",
    # inplace implementation
    "dbnd_run_start",
    "dbnd_run_stop",
    "auto_namespace",
    "namespace",
    "task_namespace",
    "as_task",
    # tasks
    "band",
    "pipeline",
    "data_source_pipeline",
    "task",
    "Task",
    # class tasks
    "Config",
    "DataSourceTask",
    "PipelineTask",
    "PythonTask",
    # parameters
    "parameter",
    "data",
    "output",
    # config
    "dbnd_config",
    "override",
    "config",
    "config_deco",
    "ConfigPath",
    "ConfigMergeSettings",
    "ParameterScope",
    "ParameterDefinition",
    # dbnd run
    "dbnd_main",
    "dbnd_cmd",
    "dbnd_run_cmd",
    "dbnd_run_cmd_main",
    "dbnd_handle_errors",
    # metrics
    "log_dataframe",
    "LogDataRequest",
    "log_metric",
    "log_metrics",
    "log_duration",
    # project paths
    "project_path",
    "relative_path",
    "databand_lib_path",
    "databand_system_path",
    # bootstrap
    "dbnd_bootstrap",
    "_set_patches",
    "track_modules",
    "track_module_functions",
    "track_functions",
]

# validate missing __all__
# imported_vars = set(k for k in locals().keys() if not k.startswith("__"))
# print(list(imported_vars.difference(set(__all__))))

# shortcuts for useful objects
str(_set_patches)  # NOQA
__version__ = "0.32.6"

__title__ = "databand"
__description__ = "Machine Learning Orchestration"
__url__ = "http://www.databand.ai/"
__uri__ = __url__
__doc__ = __description__ + " <" + __uri__ + ">"

__author__ = "Evgeny Shulman"
__email__ = "evgeny.shulman@databand.ai"

__license__ = "Commercial Licenc e"
__copyright__ = "Copyright (c) 2018 databand.ai"



import typing

from collections import OrderedDict
from typing import Optional

import attr
import six

from dbnd._core.configuration.config_value import ConfigValue


# default type of section/values container
_config_dict_type = OrderedDict
_config_id = 0


def _lower_config_name(v):
    return v.lower()


@attr.s
class _ConfigMergeSettings(object):
    on_non_exists_only = attr.ib(default=False)
    on_change_only = attr.ib(default=False)
    replace_section = attr.ib(default=False)


class ConfigMergeSettings(object):
    default = _ConfigMergeSettings()
    on_change_only = _ConfigMergeSettings(on_change_only=True)
    on_non_exists_only = _ConfigMergeSettings(on_non_exists_only=True)


# simple class that represents two level dict
class _ConfigStore(OrderedDict):
    def __init__(self, *args, **kwargs):
        OrderedDict.__init__(self, *args, **kwargs)
        self.source = None

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key.lower())
        except KeyError:
            self[key] = section = _config_dict_type()
            return section

    def get_config_value(self, section, key):
        # type: (str, str)->Optional[ConfigValue]
        section = self.get(_lower_config_name(section))
        if not section:
            return None
        return section.get(_lower_config_name(key))

    def set_config_value(self, section, key, value):
        self[_lower_config_name(section)][_lower_config_name(key)] = value

    def merge(self, config_values, merge_settings=None):
        # type : (_ConfigStore, _ConfigMergeSettings)-> _ConfigStore
        # known issue: non consisten copy on write
        # for example  c1.merge()  -> return c2.   c1.update() will affect c2 as well

        if not config_values:
            return self
        new_config = self.copy()
        # we copy only section that are changed
        for section in config_values.keys():
            new_config[section] = self[section].copy()
        new_config.update(config_values, merge_settings)
        return new_config
        # return original if no changes,
        # copy on write

    def update(self, config_values, merge_settings=None):
        # type : (_ConfigStore, _ConfigMergeSettings) -> _ConfigStore
        merge_settings = merge_settings or ConfigMergeSettings.default
        for section, section_values in six.iteritems(config_values):
            # shallow copy of configuration
            section = _lower_config_name(section)
            replace_current_section = section_values.get(
                "_replace_section", merge_settings.replace_section
            )

            if replace_current_section:
                self[section] = current_section = _config_dict_type()
            else:
                current_section = self[section]

            for key, value in six.iteritems(section_values):
                key = _lower_config_name(key)
                assert isinstance(
                    value, ConfigValue
                ), "{section}.{key} expected to be ConfigValue".format(
                    section=section, key=key
                )
                old_value = current_section.get(key)
                if old_value:
                    # if old value is override, we will "override" only if the new one is override
                    if old_value.override and not value.override:
                        continue
                    if (
                        merge_settings.on_non_exists_only
                        or value.set_if_not_exists_only
                    ):
                        continue
                    if merge_settings.on_change_only and old_value.value == value.value:
                        continue
                current_section[key] = value
        return self

    def as_value_dict(self, sections=None):
        # type: (Set[str])-> _config_dict_type
        return _config_dict_type(
            (
                section,
                _config_dict_type(
                    (key, config_value.value)
                    for key, config_value in six.iteritems(values)
                ),
            )
            for section, values in six.iteritems(self)
            if not sections or section in sections
        )


if typing.TYPE_CHECKING:
    # helping typehints system, our ordered dict is typed dict
    # not serializable in python < 3.7 (python bug)
    _TConfigStore = typing.Union[
        OrderedDict[str, OrderedDict[str, ConfigValue]], _ConfigStore
    ]



import typing

import attr

from dbnd._core.constants import MetricSource
from dbnd._core.tracking.log_data_request import LogDataRequest
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.utils.timezone import utcnow


if typing.TYPE_CHECKING:
    from typing import Any, Dict, List, Optional, Sequence, Tuple, Union


# keep it below VALUE_PREVIEW_MAX_LEN at web
_DEFAULT_VALUE_PREVIEW_MAX_LEN = 10000


@attr.s(slots=True)
class ValueMeta(object):
    value_preview = attr.ib()  # type: str
    data_dimensions = attr.ib(default=None)  # type: Optional[Sequence[int]]
    data_schema = attr.ib(default=None)  # type: Optional[Dict[str,Any]]
    data_hash = attr.ib(default=None)  # type: Optional[str]
    descriptive_stats = attr.ib(
        default=None
    )  # type: Optional[Dict[str, Dict[str, Union[int, float]]]]
    histograms = attr.ib(default=None)  # type: Optional[Dict[str, Tuple]]
    histogram_system_metrics = attr.ib(default=None)  # type: Optional[Dict]

    def build_metrics_for_key(self, key, meta_conf=None):
        # type: (str, Optional[ValueMetaConf]) -> Dict[str, List[Metric]]
        ts = utcnow()
        dataframe_metric_value = {}
        data_metrics, hist_metrics = [], []

        if self.data_dimensions:
            dataframe_metric_value["data_dimensions"] = self.data_dimensions
            for dim, size in enumerate(self.data_dimensions):
                data_metrics.append(
                    Metric(
                        key="{}.shape{}".format(key, dim),
                        value=size,
                        source=MetricSource.user,
                        timestamp=ts,
                    )
                )

        if meta_conf and meta_conf.log_schema:
            dataframe_metric_value["schema"] = self.data_schema
            data_metrics.append(
                Metric(
                    key="{}.schema".format(key),
                    value_json=self.data_schema,
                    source=MetricSource.user,
                    timestamp=ts,
                )
            )

        if meta_conf and meta_conf.log_preview:
            dataframe_metric_value["value_preview"] = self.value_preview
            dataframe_metric_value["type"] = "dataframe_metric"
            data_metrics.append(
                Metric(
                    key=str(key),
                    value_json=dataframe_metric_value,
                    source=MetricSource.user,
                    timestamp=ts,
                )
            )

        if self.histogram_system_metrics:
            hist_metrics.append(
                Metric(
                    key="{}.histogram_system_metrics".format(key),
                    value_json=self.histogram_system_metrics,
                    source=MetricSource.histograms,
                    timestamp=ts,
                )
            )
        if self.histograms:
            hist_metrics.append(
                Metric(
                    key="{}.histograms".format(key),
                    value_json=self.histograms,
                    source=MetricSource.histograms,
                    timestamp=ts,
                )
            )
        if self.descriptive_stats:
            hist_metrics.append(
                Metric(
                    key="{}.stats".format(key),
                    value_json=self.descriptive_stats,
                    source=MetricSource.histograms,
                    timestamp=ts,
                )
            )
            for column, stats in self.descriptive_stats.items():
                for stat, value in stats.items():
                    hist_metrics.append(
                        Metric(
                            key="{}.{}.{}".format(key, column, stat),
                            value=value,
                            source=MetricSource.histograms,
                            timestamp=ts,
                        )
                    )
        return {"user": data_metrics, "histograms": hist_metrics}


@attr.s
class ValueMetaConf(object):
    log_preview = attr.ib(default=None)  # type: Optional[bool]
    log_preview_size = attr.ib(default=None)  # type: Optional[int]
    log_schema = attr.ib(default=None)  # type: Optional[bool]
    log_size = attr.ib(default=None)  # type: Optional[bool]
    log_histograms = attr.ib(
        default=None, converter=LogDataRequest.from_user_param
    )  # type: Optional[Union[LogDataRequest, bool]]
    log_stats = attr.ib(
        default=None, converter=LogDataRequest.from_user_param
    )  # type: Optional[Union[LogDataRequest, bool]]

    def get_preview_size(self):
        return self.log_preview_size or _DEFAULT_VALUE_PREVIEW_MAX_LEN

    @classmethod
    def enabled(cls):
        return ValueMetaConf(
            log_size=True,
            log_preview=True,
            log_schema=True,
            log_stats=True,
            log_histograms=True,
        )



# Copyright 2009 Brian Quinlan. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.


"""
Only for python 2.7 backwards compatibility.
In python 3, this package is already included in the standard library.
"""


"""Execute computations asynchronously using threads or processes."""

__author__ = 'Brian Quinlan (brian@sweetapp.com)'

from dbnd._vendor.futures._base import (FIRST_COMPLETED,
                                      FIRST_EXCEPTION,
                                      ALL_COMPLETED,
                                      CancelledError,
                                      TimeoutError,
                                      Future,
                                      Executor,
                                      wait,
                                      as_completed)
from dbnd._vendor.futures.thread import ThreadPoolExecutor

try:
    from dbnd._vendor.process import ProcessPoolExecutor
except ImportError:
    # some platforms don't have multiprocessing
    pass



from dbnd._core.decorator.task_decorator_spec import build_task_decorator_spec
from dbnd._core.errors import TaskClassNotFoundException
from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task_build.task_metaclass import TaskMetaclass
from dbnd._core.task_build.task_registry import get_task_registry


def build_dynamic_task_class(
    task_class, task_family, user_params=None, track_source_code=False
):
    _classdict = _build_classdict(
        task_class, task_family, user_params, track_source_code
    )
    return TaskMetaclass(task_family, (task_class,), _classdict)


def _build_classdict(task_class, task_family, user_params, track_source_code):
    """
    build a classdict as needed in the creation of TaskMetaclass
    """

    attributes = dict(
        _conf__task_family=task_family,
        _conf__decorator_spec=build_task_decorator_spec(
            class_or_func=task_class,
            decorator_kwargs={},
            default_result=parameter.output.pickle[object],
        ),
        __doc__=task_class.__doc__,
        __module__=task_class.__module__,
        _conf__track_source_code=track_source_code,
    )
    if user_params:
        attributes.update(user_params)

    return attributes



import importlib
import logging

from dbnd._core.configuration import environ_config, get_dbnd_project_config
from dbnd._core.errors import friendly_error
from dbnd._core.plugin import dbnd_plugin_spec
from dbnd._core.utils.seven import import_errors
from dbnd._vendor import pluggy


logger = logging.getLogger(__name__)


hookimpl = pluggy.HookimplMarker("dbnd")
pm = pluggy.PluginManager("dbnd")
pm.add_hookspecs(dbnd_plugin_spec)

_AIRFLOW_PACKAGE_INSTALLED = None  # apache airflow is installed
_AIRFLOW_ENABLED = None  # dbnd-airflow is installed and enabled


def _is_airflow_enabled():
    if pm.has_plugin("dbnd-airflow"):
        return True

    if get_dbnd_project_config().is_no_modules:
        return False

    # TODO: make decision based on plugin only
    try:
        import dbnd_airflow

        return True
    except ImportError:
        return False


def is_airflow_enabled():
    global _AIRFLOW_ENABLED
    if _AIRFLOW_ENABLED is None:
        _AIRFLOW_ENABLED = _is_airflow_enabled()
    return _AIRFLOW_ENABLED


def disable_airflow_plugin():
    global _AIRFLOW_ENABLED
    _AIRFLOW_ENABLED = False
    pm.set_blocked("dbnd-airflow")


def use_airflow_connections():
    from dbnd._core.configuration.dbnd_config import config

    return is_airflow_enabled() and config.getboolean("airflow", "use_connections")


def assert_airflow_enabled():
    if not is_airflow_enabled():
        raise friendly_error.config.missing_module("dbnd-airflow")
    return True


def assert_airflow_package_installed():
    global _AIRFLOW_PACKAGE_INSTALLED
    if _AIRFLOW_PACKAGE_INSTALLED is None:
        try:
            import airflow

            _AIRFLOW_PACKAGE_INSTALLED = True
        except Exception:
            _AIRFLOW_PACKAGE_INSTALLED = False
    return _AIRFLOW_PACKAGE_INSTALLED


# all other modules
def is_plugin_enabled(module, module_import=None):
    if pm.has_plugin(module):
        return True

    if module_import:
        try:
            importlib.import_module(module_import)
            return True
        except import_errors:
            pass
    if "-" in module:
        # Workaround for manually loaded plugins (contain `_` instead of `-`)
        # Occurs when 'disable_pluggy_entrypoint_loading' is turned on in spark-config. All plugins are loaded
        # manually (spark_ctrl.py:107), so we must fix `plugin` syntax ('-') to python module syntax ('_')
        return is_plugin_enabled(module.replace("-", "_"), module_import)
    return False


def assert_plugin_enabled(module, reason=None, module_import=None):
    if not is_plugin_enabled(module, module_import=module_import):
        raise friendly_error.config.missing_module(module, reason)
    return True


def is_web_enabled():
    return is_plugin_enabled("dbnd-web", "dbnd_web")


def assert_web_enabled(reason=None):
    return assert_plugin_enabled("dbnd-web", reason=reason, module_import="dbnd_web")



# The moment we access this module, we need to find out if we need to use vendored protobuf package

import logging
import os
import sys

import dbnd


logger = logging.getLogger(__name__)


def is_google_installed():
    try:
        import google

        return True
    except ImportError:
        return False


def is_potobuf_installed():
    try:
        from google import protobuf

        return True
    except ImportError:
        return False


def add_vendored_package_to_syspath():
    vendor_package_str = os.path.join(os.path.dirname(dbnd.__file__), "_vendor_package")
    sys.path.append(vendor_package_str)
    logger.info("Adding `dbnd/_vendor_package/` to `sys.path`.")
    assert is_potobuf_installed()


if is_google_installed():
    from dbnd._vendor_package.google import protobuf as vendored_protobuf

    if is_potobuf_installed():
        from google import protobuf

        logger.info(
            "Using `google.protobuf` package that is available in the current python env"
        )

        if protobuf.__version__ != vendored_protobuf.__version__:
            logger.warn(
                "Excpected `protobuf==%s` but %s is installed in the current python env",
                vendored_protobuf.__version__,
                protobuf.__version__,
            )
    else:
        raise ImportError(
            "Unable to use vendored protobuf package while there are other google packages installed. "
            "Please use `pip install protobuf==3.13.0` to proceed."
        )
else:
    if dbnd.config.getboolean("core", "allow_vendored_package"):
        logger.info("Missing `google.protobuf` package in the current python env.")
        add_vendored_package_to_syspath()
    else:
        raise ImportError(
            "Missing `google.protobuf` package in the current python env while "
            "core.allow_vendored_package=False"
        )



from dbnd._core.errors import DatabandConfigError, DatabandRuntimeError
from dbnd._core.errors.friendly_error.helpers import _run_name


def dataflow_pipeline_not_set(task):
    return DatabandRuntimeError(
        "dataflow_pipeline at {task} is None. Can't wait on dataflow job completion.".format(
            task=_run_name(task)
        ),
        help_msg="Please set task.pipeline first at '{task}' or change task.dataflow_wait_until_finish flag".format(
            task=_run_name(task)
        ),
    )



from typing import Type

from dbnd._core.parameter.parameter_builder import parameter
from dbnd._core.task import Config
from targets.target_config import TargetConfig, file, parse_target_config
from targets.values import get_value_type_of_type
from targets.values.version_value import VersionAlias, VersionStr


class OutputConfig(Config):
    """(Advanced) Databand's core task's output behaviour"""

    _conf__task_family = "output"

    path_task = parameter(description="default path for every Task")[str]
    path_prod_immutable_task = parameter(
        description="format of the path to be used by Production Immutable tasks"
    )[str]

    hdf_format = (
        parameter.choices(["table", "fixed"])
        .help("Default format to save DataFrame to hdf")
        .value("fixed")
    )

    deploy_id = parameter(
        default=VersionAlias.context_uid,
        description="deploy prefix to use for remote deployments",
    )[VersionStr]

    def get_config(self, value_type):
        # type: (Type) -> TargetConfig

        type_handler = get_value_type_of_type(value_type)
        for possible_option in [str(type_handler), type_handler.config_name]:
            config_value = self.task_meta.get_task_config_value(key=possible_option)
            if config_value:
                return parse_target_config(config_value.value)
        return file.pickle



from collections import namedtuple

from dbnd._core.current import get_databand_context
from dbnd._core.task_build.task_context import has_current_task


CONF_TASK_ENV_SECTION = "task_env"
CONF_TASK_SECTION = "task"
CONF_CONFIG_SECTION = "config"

ConfigPath = namedtuple("ConfigPath", ["section", "key"])


def from_task_env(key):
    return ConfigPath(CONF_TASK_ENV_SECTION, key)


def from_current_env(key):
    if has_current_task():
        return from_task_env(key)

    return ConfigPath(get_databand_context().env.task_name, key)



import os

from dbnd import get_databand_context
from dbnd.tasks.py_distribution.build_distribution import (
    build_fat_requirements_py_zip_file,
    build_package_zip,
    build_third_party_zip,
)
from dbnd.tasks.py_distribution.fat_wheel_config import FatWheelConfig


_bdist_zip_cache = {}
_bdist_zip_list_cache = {}


def build_fat_wheel(use_cached=True, output_dir=None):
    """
    Builds wheel files for package and third party requirements (if given)
    and merges them into one big wheel file (with zip suffix).
    Returns the path to the created zip file.
    """
    output_dir = output_dir or _get_output_dir(use_cached=use_cached)
    conf = FatWheelConfig()
    bdist_zip_cache_key = (conf.package_dir, conf.requirements_file)

    if (
        use_cached
        and os.path.exists(output_dir.path)
        and bdist_zip_cache_key in _bdist_zip_cache
    ):
        return _bdist_zip_cache[bdist_zip_cache_key]

    result_file = build_fat_requirements_py_zip_file(
        output_dir, conf.package_dir, conf.requirements_file
    )
    _bdist_zip_cache[bdist_zip_cache_key] = result_file
    return result_file


def build_wheel_zips(use_cached=True, output_dir=None):
    """
    Builds one wheel file for the package (with zip suffix) and one wheel file
    for all third party requirements (zip suffix as well).
    Returns a list that contains the paths of the two zip files.
    """
    output_dir = output_dir or _get_output_dir(use_cached=use_cached)
    conf = FatWheelConfig()
    bdist_zip_cache_key = (conf.package_dir, conf.requirements_file)
    if (
        use_cached
        and os.path.exists(output_dir.path)
        and bdist_zip_cache_key in _bdist_zip_cache
    ):
        return _bdist_zip_list_cache[bdist_zip_cache_key]

    package_zip = build_package_zip(output_dir, conf.package_dir)
    result_file_list = [package_zip]
    if conf.requirements_file:
        third_party_zip = build_third_party_zip(output_dir, conf.requirements_file)
        result_file_list.append(third_party_zip)
    _bdist_zip_list_cache[bdist_zip_cache_key] = result_file_list
    return result_file_list


def _get_output_dir(use_cached):
    dc = get_databand_context()
    working_dir = dc.env.dbnd_local_root__build
    if use_cached:
        working_dir = working_dir.folder("bdist_zip_build_%s" % dc.current_context_uid)
    else:
        import random

        random_number = random.randrange(0, 1000)
        working_dir = working_dir.folder(
            "bdist_zip_build_%s_%d" % (dc.current_context_uid, random_number)
        )
    return working_dir



from dbnd._core.current import get_databand_context
from dbnd.api.shared_schemas.airflow_monitor import AirflowServerInfoSchema


def list_synced_airflow_instances():
    endpoint = "airflow_monitor"
    schema = AirflowServerInfoSchema(strict=False)
    client = get_databand_context().databand_api_client
    response = client.api_request(endpoint=endpoint, data="", method="GET")
    return schema.load(data=response["data"], many=True).data


def create_airflow_instance(url, external_url, fetcher, api_mode, composer_client_id):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/add"
    request_data = {
        "base_url": url,
        "ext_url": external_url,
        "fetcher": fetcher,
        "api_mode": api_mode,
        "composer_client_id": composer_client_id,
    }
    client.api_request(endpoint, request_data, method="POST")


def archive_airflow_instance(url):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/archive"
    client.api_request(endpoint, url, method="POST")


def unarchive_airflow_instance(url):
    client = get_databand_context().databand_api_client
    endpoint = "airflow_monitor/unarchive"
    client.api_request(endpoint, url, method="POST")



from typing import Any


class _Nothing(object):
    """
    based on Attrs version
    Sentinel class to indicate the lack of a value when ``None`` is ambiguous.

    All instances of `_Nothing` are equal.
    """

    def __copy__(self):
        return self

    def __deepcopy__(self, _):
        return self

    def __eq__(self, other):
        return other.__class__ == _Nothing

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "NOTHING"

    def __hash__(self):
        return 0xC0FFEE

    def __bool__(self):
        return False

    def __len__(self):
        return 0

    def __nonzero__(self):
        return False


NOTHING = _Nothing()


def is_defined(value):  # type: (Any) -> bool
    if value is None:
        return True  # value defined as None

    # we can not check with   'value is NOTHING' as NOTHING ref is changed after serialization/deserialization
    if isinstance(value, _Nothing):
        return False

    return True


def is_not_defined(value):
    return not is_defined(value)


def get_or_default(value, default):
    if is_defined(value):
        return value
    return default



from typing import Any, Dict, Iterable, List, Tuple, Type

from dbnd._core.parameter.parameter_definition import ParameterDefinition
from dbnd._core.parameter.parameter_value import ParameterValue
from dbnd._core.task_ctrl.task_meta import TaskMeta


class TaskParameters(object):
    def __init__(self, task):
        self.task = task
        self.task_meta = self.task.task_meta  # type: TaskMeta
        self._params = [
            p_value.parameter for p_value in self.task_meta.class_task_params
        ]
        self._param_obj_map = {p.name: p for p in self._params}

    def get_param(self, param_name):
        return self._param_obj_map.get(param_name, None)

    def get_value(self, param_name):
        # we dont' want to autoread when we run this function
        # most cases we are running it from "banner" print
        # so if we are in the autoread we will use original values
        task_auto_read_original = self.task._task_auto_read_original
        if task_auto_read_original is not None:
            return task_auto_read_original[param_name]

        return getattr(self.task, param_name)

    def get_param_meta(self, param_name):  # type: (str) -> ParameterValue
        return self.task_meta._task_params.get_any_param(param_name, None)

    def get_params(
        self,
        param_cls=ParameterDefinition,
        significant_only=False,
        input_only=False,
        output_only=False,
        user_only=False,
    ):
        # type: (...)-> List[ParameterDefinition]
        result = self._params
        if param_cls is not None:
            result = filter(lambda p: isinstance(p, param_cls), result)
        if significant_only:
            result = filter(lambda p: p.significant, result)
        if input_only:
            result = filter(lambda p: not p.is_output(), result)
        if output_only:
            result = filter(lambda p: p.is_output(), result)
        if user_only:
            result = filter(lambda p: not p.system, result)
        return list(result)

    def get_param_values(
        self,
        param_cls=ParameterDefinition,
        significant_only=False,
        input_only=False,
        output_only=False,
        user_only=False,
    ):
        # type: (Type[ParameterDefinition], bool, bool, bool, bool) -> List[ Tuple[ParameterDefinition, Any]]
        result = self.get_params(
            param_cls=param_cls,
            significant_only=significant_only,
            input_only=input_only,
            output_only=output_only,
            user_only=user_only,
        )

        result = [(p, self.get_value(p.name)) for p in result]
        return result

    # TODO: change name to "to_string"
    def get_params_serialized(
        self, param_cls=ParameterDefinition, significant_only=False, input_only=False
    ):
        return [
            (p.name, p.signature(value))
            for p, value in self.get_param_values(
                param_cls=param_cls,
                significant_only=significant_only,
                input_only=input_only,
            )
        ]

    def to_env_map(self, *param_names):
        # type: (List[str]) -> Dict[str, str]
        return {
            p.get_env_key(self.task.task_name): p.to_str(value)
            for p, value in self.get_param_values()
            if value is not None and (len(param_names) == 0 or p.name in param_names)
        }

    def get_param_env_key(self, param_name):
        return self.get_param(param_name).get_env_key(self.task.task_name)

    def __iter__(self):
        # type: (TaskParameters) -> Iterable[ParameterDefinition, ParameterValue, ParameterValue]
        """helper function to iterate all the params with it's definition, value and meta"""

        for definition in self.get_params():
            value = self.get_value(definition.name)
            meta = self.get_param_meta(definition.name)
            yield definition, value, meta



import json

import attr

from dbnd._core.constants import _DbndDataClass


class Metric(_DbndDataClass):
    _int_precision = 16
    # Workaround for precission issues with large ints
    MAX_INT = 10 ** (_int_precision - 1)
    MIN_INT = -MAX_INT

    def __init__(
        self,
        key,
        timestamp,
        value_int=None,
        value_float=None,
        value_str=None,
        value_json=None,
        value=None,
        source=None,
    ):
        self.key = key
        self.timestamp = timestamp
        self.source = source
        self.value_int = value_int
        self.value_float = value_float
        self.value_str = value_str
        self.value_json = None
        if value_json:
            self.value_json = value_json
            self.value_str = None
        elif not (value_int or value_float or value_str):
            self.value = value

    @property
    def value(self):
        if self.value_float is not None:
            return self.value_float
        if self.value_int is not None:
            return int(self.value_int)
        if self.value_json is not None:
            return self.value_json
        return self.value_str

    @property
    def serialized_value(self):
        if self.value_json is not None:
            return json.dumps(self.value_json, sort_keys=True)
        else:
            return str(self.value)

    @value.setter
    def value(self, value):
        if isinstance(value, float):
            self.value_float = value
        elif isinstance(value, int) and self.MIN_INT <= value <= self.MAX_INT:
            self.value_int = value
        else:
            self.value_str = value

    def __repr__(self):
        return "Metric(key={}, source={})".format(self.key, self.source)


@attr.s
class Artifact(_DbndDataClass):
    path = attr.ib()  # type: str



import sys
import typing


class PathStr(str):
    pass


if sys.version_info > (3, 5):  # pragma: nocover
    # noinspection PyUnresolvedReferences
    import pathlib

    # noinspection PyUnresolvedReferences
    from pathlib import Path
else:  # pragma: nocover
    # noinspection PyUnresolvedReferences
    import pathlib2 as pathlib

    # noinspection PyUnresolvedReferences
    from pathlib2 import Path


class NullableStr(str):
    pass


# Data typing
class LazyLoad(object):
    pass


class EagerLoad(object):
    pass


T = typing.TypeVar("T")
KT = typing.TypeVar("KT")
VT = typing.TypeVar("VT")


class DataList(typing.List[T], LazyLoad):
    pass


class DataDict(typing.Dict[KT, VT], LazyLoad):
    pass



import pandas as pd

from dbnd import data, output
from dbnd.tasks import PythonTask


class PandasFrameToParquet(PythonTask):
    source = data[pd.DataFrame]
    dest = output.pandas_dataframe.parquet

    def run(self):
        self.dest = self.source



import logging

from dbnd._core.constants import TaskRunState
from dbnd._core.errors.base import DatabandRunError, DatabandSigTermError
from dbnd._core.task_ctrl.task_dag import topological_sort
from dbnd._core.task_executor.task_executor import TaskExecutor


logger = logging.getLogger(__name__)


class LocalTaskExecutor(TaskExecutor):
    def do_run(self):
        topological_tasks = topological_sort([tr.task for tr in self.task_runs])
        fail_fast = self.settings.run.fail_fast
        task_failed = False

        task_runs_to_update_state = []
        for task in topological_tasks:
            tr = self.run.get_task_run_by_id(task.task_id)
            if tr.is_reused:
                continue

            if fail_fast and task_failed:
                logger.info(
                    "Setting %s to %s", task.task_id, TaskRunState.UPSTREAM_FAILED
                )
                tr.set_task_run_state(TaskRunState.UPSTREAM_FAILED, track=False)
                task_runs_to_update_state.append(tr)
                continue

            upstream_task_runs = [
                self.run.get_task_run_by_id(t.task_id)
                for t in task.ctrl.task_dag.upstream
            ]
            failed_upstream = [
                upstream_task_run
                for upstream_task_run in upstream_task_runs
                if upstream_task_run.task_run_state in TaskRunState.fail_states()
            ]
            if failed_upstream:
                logger.info(
                    "Setting %s to %s", task.task_id, TaskRunState.UPSTREAM_FAILED
                )
                tr.set_task_run_state(TaskRunState.UPSTREAM_FAILED, track=False)
                task_runs_to_update_state.append(tr)
                continue

            if self.run.is_killed():
                logger.info(
                    "Databand Context is killed! Stopping %s to %s",
                    task.task_id,
                    TaskRunState.FAILED,
                )
                tr.set_task_run_state(TaskRunState.FAILED, track=False)
                task_runs_to_update_state.append(tr)
                continue

            logger.debug("Executing task: %s", task.task_id)

            try:
                tr.runner.execute()
            except DatabandSigTermError as e:
                raise e
            except Exception as e:
                task_failed = True
                logger.error("Failed to execute task '%s': %s" % (task.task_id, str(e)))

        if task_runs_to_update_state:
            self.run.tracker.set_task_run_states(task_runs_to_update_state)

        if task_failed:
            err = _collect_errors(self.run.task_runs)

            if err:
                raise DatabandRunError(err)


def _collect_errors(task_runs):
    err = ""
    upstream_failed = []
    failed = []
    for task_run in task_runs:
        task_name = task_run.task.task_name
        if task_run.task_run_state == TaskRunState.UPSTREAM_FAILED:
            # we don't want to show upstream failed in the list
            upstream_failed.append(task_name)
        elif task_run.task_run_state in TaskRunState.direct_fail_states():
            failed.append(task_name)
    if upstream_failed:
        err += "Task that didn't run because of failed dependency:\n\t{}\n".format(
            "\n\t".join(upstream_failed)
        )
    if failed:
        err += "Failed tasks are:\n\t{}".format("\n\t".join(failed))
    return err



# -*- coding: utf-8 -*-
#
# Copyright 2012-2015 Spotify AB
# Modifications copyright (C) 2018 databand.ai
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


def get_editdistance(a, b):
    """ Simple unweighted Levenshtein distance """
    r0 = range(0, len(b) + 1)
    r1 = [0] * (len(b) + 1)

    for i in range(0, len(a)):
        r1[0] = i + 1

        for j in range(0, len(b)):
            c = 0 if a[i] is b[j] else 1
            r1[j + 1] = min(r1[j] + 1, r0[j + 1] + 1, r0[j] + c)

        r0 = r1[:]

    return r1[len(b)]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tracking.proto
"""Generated protocol buffer code."""
from google.protobuf import (
    descriptor as _descriptor,
    message as _message,
    reflection as _reflection,
    struct_pb2 as google_dot_protobuf_dot_struct__pb2,
    symbol_database as _symbol_database,
    timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2,
)


# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="tracking.proto",
    package="dbnd.services.tracking",
    syntax="proto3",
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_pb=b'\n\x0etracking.proto\x12\x16\x64\x62nd.services.tracking\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\xe5\x01\n\x05\x45vent\x12\x0c\n\x04uuid\x18\x01 \x01(\t\x12\x0e\n\x06schema\x18\x02 \x01(\t\x12%\n\x04\x64\x61ta\x18\x03 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x39\n\x06labels\x18\x04 \x03(\x0b\x32).dbnd.services.tracking.Event.LabelsEntry\x12-\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a-\n\x0bLabelsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"q\n\x11PostEventsRequest\x12-\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1d.dbnd.services.tracking.Event\x12-\n\ttimestamp\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"S\n\x12PostEventsResponse\x12*\n\tresponses\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x11\n\texception\x18\x02 \x01(\t"s\n\x12SearchEventRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12\x16\n\tpage_size\x18\x02 \x01(\rH\x00\x88\x01\x01\x12\x18\n\x0bpage_number\x18\x03 \x01(\rH\x01\x88\x01\x01\x42\x0c\n\n_page_sizeB\x0e\n\x0c_page_number"D\n\x13SearchEventResponse\x12-\n\x06\x65vents\x18\x01 \x03(\x0b\x32\x1d.dbnd.services.tracking.Event2\xc8\x02\n\x0fTrackingService\x12\x63\n\nPostEvents\x12).dbnd.services.tracking.PostEventsRequest\x1a*.dbnd.services.tracking.PostEventsResponse\x12g\n\x0cStreamEvents\x12).dbnd.services.tracking.PostEventsRequest\x1a*.dbnd.services.tracking.PostEventsResponse(\x01\x12g\n\x0cSearchEvents\x12*.dbnd.services.tracking.SearchEventRequest\x1a+.dbnd.services.tracking.SearchEventResponseb\x06proto3',
    dependencies=[
        google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,
        google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,
    ],
)


_EVENT_LABELSENTRY = _descriptor.Descriptor(
    name="LabelsEntry",
    full_name="dbnd.services.tracking.Event.LabelsEntry",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="key",
            full_name="dbnd.services.tracking.Event.LabelsEntry.key",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="value",
            full_name="dbnd.services.tracking.Event.LabelsEntry.value",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=b"8\001",
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=290,
    serialized_end=335,
)

_EVENT = _descriptor.Descriptor(
    name="Event",
    full_name="dbnd.services.tracking.Event",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="uuid",
            full_name="dbnd.services.tracking.Event.uuid",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="schema",
            full_name="dbnd.services.tracking.Event.schema",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="data",
            full_name="dbnd.services.tracking.Event.data",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="labels",
            full_name="dbnd.services.tracking.Event.labels",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="dbnd.services.tracking.Event.timestamp",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[_EVENT_LABELSENTRY,],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=106,
    serialized_end=335,
)


_POSTEVENTSREQUEST = _descriptor.Descriptor(
    name="PostEventsRequest",
    full_name="dbnd.services.tracking.PostEventsRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="events",
            full_name="dbnd.services.tracking.PostEventsRequest.events",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="timestamp",
            full_name="dbnd.services.tracking.PostEventsRequest.timestamp",
            index=1,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=337,
    serialized_end=450,
)


_POSTEVENTSRESPONSE = _descriptor.Descriptor(
    name="PostEventsResponse",
    full_name="dbnd.services.tracking.PostEventsResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="responses",
            full_name="dbnd.services.tracking.PostEventsResponse.responses",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="exception",
            full_name="dbnd.services.tracking.PostEventsResponse.exception",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=452,
    serialized_end=535,
)


_SEARCHEVENTREQUEST = _descriptor.Descriptor(
    name="SearchEventRequest",
    full_name="dbnd.services.tracking.SearchEventRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="query",
            full_name="dbnd.services.tracking.SearchEventRequest.query",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=b"".decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="page_size",
            full_name="dbnd.services.tracking.SearchEventRequest.page_size",
            index=1,
            number=2,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.FieldDescriptor(
            name="page_number",
            full_name="dbnd.services.tracking.SearchEventRequest.page_number",
            index=2,
            number=3,
            type=13,
            cpp_type=3,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="_page_size",
            full_name="dbnd.services.tracking.SearchEventRequest._page_size",
            index=0,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
        _descriptor.OneofDescriptor(
            name="_page_number",
            full_name="dbnd.services.tracking.SearchEventRequest._page_number",
            index=1,
            containing_type=None,
            create_key=_descriptor._internal_create_key,
            fields=[],
        ),
    ],
    serialized_start=537,
    serialized_end=652,
)


_SEARCHEVENTRESPONSE = _descriptor.Descriptor(
    name="SearchEventResponse",
    full_name="dbnd.services.tracking.SearchEventResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    create_key=_descriptor._internal_create_key,
    fields=[
        _descriptor.FieldDescriptor(
            name="events",
            full_name="dbnd.services.tracking.SearchEventResponse.events",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            serialized_options=None,
            file=DESCRIPTOR,
            create_key=_descriptor._internal_create_key,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    serialized_options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=654,
    serialized_end=722,
)

_EVENT_LABELSENTRY.containing_type = _EVENT
_EVENT.fields_by_name["data"].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_EVENT.fields_by_name["labels"].message_type = _EVENT_LABELSENTRY
_EVENT.fields_by_name[
    "timestamp"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_POSTEVENTSREQUEST.fields_by_name["events"].message_type = _EVENT
_POSTEVENTSREQUEST.fields_by_name[
    "timestamp"
].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_POSTEVENTSRESPONSE.fields_by_name[
    "responses"
].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_SEARCHEVENTREQUEST.oneofs_by_name["_page_size"].fields.append(
    _SEARCHEVENTREQUEST.fields_by_name["page_size"]
)
_SEARCHEVENTREQUEST.fields_by_name[
    "page_size"
].containing_oneof = _SEARCHEVENTREQUEST.oneofs_by_name["_page_size"]
_SEARCHEVENTREQUEST.oneofs_by_name["_page_number"].fields.append(
    _SEARCHEVENTREQUEST.fields_by_name["page_number"]
)
_SEARCHEVENTREQUEST.fields_by_name[
    "page_number"
].containing_oneof = _SEARCHEVENTREQUEST.oneofs_by_name["_page_number"]
_SEARCHEVENTRESPONSE.fields_by_name["events"].message_type = _EVENT
DESCRIPTOR.message_types_by_name["Event"] = _EVENT
DESCRIPTOR.message_types_by_name["PostEventsRequest"] = _POSTEVENTSREQUEST
DESCRIPTOR.message_types_by_name["PostEventsResponse"] = _POSTEVENTSRESPONSE
DESCRIPTOR.message_types_by_name["SearchEventRequest"] = _SEARCHEVENTREQUEST
DESCRIPTOR.message_types_by_name["SearchEventResponse"] = _SEARCHEVENTRESPONSE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Event = _reflection.GeneratedProtocolMessageType(
    "Event",
    (_message.Message,),
    {
        "LabelsEntry": _reflection.GeneratedProtocolMessageType(
            "LabelsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _EVENT_LABELSENTRY,
                "__module__": "tracking_pb2"
                # @@protoc_insertion_point(class_scope:dbnd.services.tracking.Event.LabelsEntry)
            },
        ),
        "DESCRIPTOR": _EVENT,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.Event)
    },
)
_sym_db.RegisterMessage(Event)
_sym_db.RegisterMessage(Event.LabelsEntry)

PostEventsRequest = _reflection.GeneratedProtocolMessageType(
    "PostEventsRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTEVENTSREQUEST,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.PostEventsRequest)
    },
)
_sym_db.RegisterMessage(PostEventsRequest)

PostEventsResponse = _reflection.GeneratedProtocolMessageType(
    "PostEventsResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _POSTEVENTSRESPONSE,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.PostEventsResponse)
    },
)
_sym_db.RegisterMessage(PostEventsResponse)

SearchEventRequest = _reflection.GeneratedProtocolMessageType(
    "SearchEventRequest",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHEVENTREQUEST,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.SearchEventRequest)
    },
)
_sym_db.RegisterMessage(SearchEventRequest)

SearchEventResponse = _reflection.GeneratedProtocolMessageType(
    "SearchEventResponse",
    (_message.Message,),
    {
        "DESCRIPTOR": _SEARCHEVENTRESPONSE,
        "__module__": "tracking_pb2"
        # @@protoc_insertion_point(class_scope:dbnd.services.tracking.SearchEventResponse)
    },
)
_sym_db.RegisterMessage(SearchEventResponse)


_EVENT_LABELSENTRY._options = None

_TRACKINGSERVICE = _descriptor.ServiceDescriptor(
    name="TrackingService",
    full_name="dbnd.services.tracking.TrackingService",
    file=DESCRIPTOR,
    index=0,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
    serialized_start=725,
    serialized_end=1053,
    methods=[
        _descriptor.MethodDescriptor(
            name="PostEvents",
            full_name="dbnd.services.tracking.TrackingService.PostEvents",
            index=0,
            containing_service=None,
            input_type=_POSTEVENTSREQUEST,
            output_type=_POSTEVENTSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="StreamEvents",
            full_name="dbnd.services.tracking.TrackingService.StreamEvents",
            index=1,
            containing_service=None,
            input_type=_POSTEVENTSREQUEST,
            output_type=_POSTEVENTSRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
        _descriptor.MethodDescriptor(
            name="SearchEvents",
            full_name="dbnd.services.tracking.TrackingService.SearchEvents",
            index=2,
            containing_service=None,
            input_type=_SEARCHEVENTREQUEST,
            output_type=_SEARCHEVENTRESPONSE,
            serialized_options=None,
            create_key=_descriptor._internal_create_key,
        ),
    ],
)
_sym_db.RegisterServiceDescriptor(_TRACKINGSERVICE)

DESCRIPTOR.services_by_name["TrackingService"] = _TRACKINGSERVICE

# @@protoc_insertion_point(module_scope)



from dbnd._core.tracking.schemas.base import ApiObjectSchema
from dbnd._core.tracking.schemas.metrics import Metric
from dbnd._core.tracking.schemas.tracking_info_objects import ErrorInfo, TargetInfo
from dbnd._vendor.marshmallow import fields, post_load


class TargetInfoSchema(ApiObjectSchema):
    parameter_name = fields.String()

    path = fields.String()
    created_date = fields.DateTime(allow_none=True)
    task_run_uid = fields.UUID(allow_none=True)

    @post_load
    def make_target(self, data, **kwargs):
        return TargetInfo(**data)


class MetricSchema(ApiObjectSchema):
    key = fields.String()
    value = fields.Raw(allow_none=True)
    value_str = fields.String(allow_none=True)
    value_json = fields.Dict(allow_none=True)
    value_int = fields.Integer(allow_none=True)
    value_float = fields.Float(allow_none=True)
    timestamp = fields.DateTime()

    @post_load
    def make_object(self, data, **kwargs):
        return Metric(**data)


class ArtifactSchema(ApiObjectSchema):
    path = fields.String()


class ErrorInfoSchema(ApiObjectSchema):
    msg = fields.String()
    help_msg = fields.String(allow_none=True)
    databand_error = fields.Bool()
    exc_type = fields.Function(lambda obj: str(obj.exc_type), allow_none=True)
    traceback = fields.String()
    nested = fields.String(allow_none=True)
    user_code_traceback = fields.String()
    show_exc_info = fields.Bool()

    @post_load
    def make_object(self, data, **kwargs):
        return ErrorInfo(**data)


class ExternalUrlSchema(ApiObjectSchema):
    id = fields.UUID(attribute="uid")
    name = fields.Str()
    url = fields.Str()



import os

from contextlib import contextmanager

import six

from targets import pipes
from targets.base_target import logger
from targets.data_target import DataTarget
from targets.errors import FileAlreadyExists, TargetError
from targets.fs import get_file_system, get_file_system_name
from targets.pipes import Nop, Text
from targets.target_config import FileCompressions, TargetConfig


class FileTarget(DataTarget):
    """
    Base class for FileSystem Targets like  :class:`~databand.contrib.hdfs.HdfsTarget`.

    A FileTarget has an associated :py:class:`FileSystem` to which certain operations can be
    delegated. By default, :py:meth:`exists` and :py:meth:`remove` are delegated to the
    :py:class:`FileSystem`, which is determined by the :py:attr:`fs` property.

    Methods of FileTarget raise :py:class:`FileSystemException` if there is a problem
    completing the operation.

    """

    def __init__(
        self, path, fs, config=None, io_pipe=None, properties=None, source=None
    ):
        """
        Initializes a FileTarget instance.

        :param str path: the path associated with this FileTarget.
        """
        super(FileTarget, self).__init__(properties=properties, source=source)
        self.path = path
        self._fs = fs
        self.fs_name = self._fs.name if self._fs else get_file_system_name(path)
        self.io_pipe = io_pipe
        self.config = config  # type: TargetConfig
        if not path:
            raise TargetError("path must be set")

    @property
    def fs(self):
        if self._fs is None:
            return get_file_system(self.fs_name)
        return self._fs

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.path == other.path

    def _get_pipe(self, mode):
        if self.io_pipe:
            return self.io_pipe

        if six.PY2:
            pipe = Nop
        # python3 has differences between binary and non binary stream, so we need to have a proper handling for that
        elif "b" not in mode:
            pipe = Text
        else:
            pipe = Nop
        if not self.config:
            return pipe

        if self.config.compression == FileCompressions.gzip:
            pipe = pipe >> pipes.Gzip
        elif self.config.compression == FileCompressions.bzip:
            pipe = pipe >> pipes.Bzip2
        return pipe

    def open(self, mode="r"):
        """
        Open the FileSystem target.

        This method returns a file-like object which can either be read from or written to depending
        on the specified mode.

        :param str mode: the mode `r` opens the FileTarget in read-only mode, whereas `w` will
                         open the FileTarget in write mode. Subclasses can implement
                         additional options.
        """

        io_pipe = self._get_pipe(mode)
        if "r" in mode:
            return io_pipe.pipe_reader(self.fs.open_read(self.path, mode=mode))
        elif "w" in mode:
            overwrite = False
            if self.config and self.config.overwrite_target:
                overwrite = True
            return io_pipe.pipe_writer(
                self.fs.open_write(self.path, mode=mode, overwrite=overwrite)
            )
        else:
            raise ValueError("Unsupported open mode '{}'".format(mode))

    def exists(self):
        """
        Returns ``True`` if the path for this FileTarget exists; ``False`` otherwise.

        This method is implemented by using :py:attr:`fs`.
        """
        path = self.path
        original_path = path
        while "*" in path or "?" in path or "[" in path or "{" in path:
            path = os.path.dirname(path)

        if path != original_path:
            logger.info(
                "Wildcard in path '%s', checking only for '%s'", original_path, path
            )
            # treat it as directory (can have different behaviour on key-values stores like s3)
            if not path.endswith(os.path.sep):
                path += os.path.sep
        return self.fs.exists(path)

    def exist_after_write_consistent(self):
        return self.fs.exist_after_write_consistent()

    def remove(self):
        """
        Remove the resource at the path specified by this FileTarget.

        This method is implemented by using :py:attr:`fs`.
        """
        self.fs.remove(self.path)

    def __repr__(self):
        return repr(self.path)

    def __str__(self):
        return str(self.path)

    # basic file ops
    def mkdir_parent(self):
        self.fs.mkdir_parent(self.path)

    def move(self, new_path, raise_if_exists=False):
        self.fs.move(self.path, str(new_path), raise_if_exists=raise_if_exists)

    def move_from(self, from_path, raise_if_exists=False):
        self.fs.move(from_path, self.path, raise_if_exists=raise_if_exists)
        self.mark_success()

    def copy_from_local(self, local_path):
        self.fs.copy_from_local(local_path, self.path)

    def download(self, local_path, **kwargs):
        self.fs.download(self.path, local_path, **kwargs)

    def copy(self, new_path, raise_if_exists=False):
        self.fs.copy(self.path, str(new_path), raise_if_exists)

    def __fspath__(self):
        """
        https://www.python.org/dev/peps/pep-0519/
        make pandas be able to get targets
        :return:
        """
        return self.path

    def mark_success(self):
        if not self.exists():
            logger.info("Target is not exists, it can not be marked as success")

    def __hash__(self):
        return hash(self.path)

    def make_tmp(self):
        return self.fs.make_tmp()

    @contextmanager
    def tmp(self):
        """
        Create a tmp path for usage and move the content from tmp to the local target at the end.
        """
        tmp_path = self.make_tmp()

        yield tmp_path

        try:
            self.move_from(tmp_path)

        except FileAlreadyExists as e:
            logger.warning(
                "Moving from %s to %s failed. File already exist! | Error: %s"
                % (tmp_path, self.path, e)
            )
        finally:
            if self.fs.exists(tmp_path):
                self.fs.remove(tmp_path)


FileSystemTarget = FileTarget



# ORIGIN: https://github.com/databricks/mlflow
import os
import shutil
import tempfile


class TempDir(object):
    def __init__(self, chdr=False, remove_on_exit=True):
        self._dir = None
        self._path = None
        self._chdr = chdr
        self._remove = remove_on_exit

    def __enter__(self):
        self._path = os.path.abspath(tempfile.mkdtemp())
        assert os.path.exists(self._path)
        if self._chdr:
            self._dir = os.path.abspath(os.getcwd())
            os.chdir(self._path)
        return self

    def __exit__(self, tp, val, traceback):
        if self._chdr and self._dir:
            os.chdir(self._dir)
            self._dir = None
        if self._remove and os.path.exists(self._path):
            shutil.rmtree(self._path)

        assert not self._remove or not os.path.exists(self._path)
        assert os.path.exists(os.getcwd())

    def path(self, *path):
        return (
            os.path.join("./", *path) if self._chdr else os.path.join(self._path, *path)
        )



import logging

from typing import Dict, List

from dbnd._core.constants import CloudType
from dbnd._core.log import dbnd_log_debug
from dbnd._core.parameter import PARAMETER_FACTORY as parameter
from dbnd._core.task import Config
from dbnd._core.utils.function_args import first_not_none
from targets import Target
from targets.value_meta import _DEFAULT_VALUE_PREVIEW_MAX_LEN, ValueMetaConf
from targets.values import ValueType


logger = logging.getLogger()


class DatabandSystemConfig(Config):
    """Databand's command line arguments (see `dbnd run --help`)"""

    _conf__task_family = "databand"

    verbose = parameter(description="Make logging output more verbose").value(False)
    print_task_band = parameter(description="Print task_band in logging output.").value(
        False
    )

    describe = parameter.help("Describe current run").value(False)

    module = parameter(
        default=None, description="Auto load this module before resolving user classes"
    )[str]
    env = parameter(
        default=CloudType.local,
        description="task environment: local/aws/aws_prod/gcp/prod",
    )[str]

    conf_file = parameter(default=None, description="List of files to read from")[
        List[str]
    ]
    # keep it last one, parameter keyword is used for parameters generation
    # all following params will be dict
    conf = parameter(
        default=None,
        description="JSON string/key=value that gets into the Task attribute",
    )[Dict[str, str]]

    project_name = parameter(
        default="databand_project", description="Name of this databand project"
    )[str]


class CoreConfig(Config):
    """Databand's core functionality behaviour"""

    _conf__task_family = "core"

    dags_subdir = parameter.system(
        default="",
        description="File location or directory from which to look for the dag",
    )

    environments = parameter(description="List of enabled environments")[List[str]]

    dbnd_user = parameter(description="user used to connect to the dbnd web server")[
        str
    ]
    dbnd_password = parameter(
        description="password used to connect to the dbnd web server"
    )[str]
    databand_url = parameter(
        default=None,
        description="Tracker URL to be used for creating links in console logs",
    )[str]
    databand_access_token = parameter(
        description="Personall access token to connect to the dbnd web server",
        default=None,
    )[str]

    # Backward compatibility
    tracker_url = parameter(
        default=None,
        description="OLD: Tracker URL to be used for creating links in console logs",
    )[str]

    tracker_version = parameter[str]

    user_configs = parameter(
        empty_default=True,
        description="Contains the config for creating tasks from user code",
    )[List[str]]

    # user_pre_init = defined at Databand System config, dbnd_on_pre_init_context
    user_init = parameter(
        default=None,
        description="Runs in every DBND process with System configuration in place (dbnd_post_enter_context)",
    )[object]
    user_driver_init = parameter(
        default=None,
        description="Runs in driver after config initialization (dbnd_on_new_context)",
    )[object]

    user_code_on_fork = parameter(
        default=None, description="Runs in sub process (parallel/kubernets/external)"
    )[object]

    pickle_handler = parameter(
        default=None,
        description="Defines a python pickle handler to be used to pickle the "
        "run's data",
    )[str]

    tracker = parameter(
        default=["file", "console", "api"], description="Tracking Stores to be used"
    )[List[str]]
    tracker_raise_on_error = parameter(
        default=True, description="Raise error when failed to track data"
    )[bool]
    tracker_api = parameter(default="web", description="Tracking Stores to be used")[
        str
    ]

    always_save_pipeline = parameter(
        description="Boolean for always saving pipeline to pickle"
    ).value(False)
    disable_save_pipeline = parameter(
        description="Boolean for disabling pipeline pickling"
    ).value(False)

    recheck_circle_dependencies = parameter(
        description="Re check circle dependencies on every task creation,"
        " use it if you need to find of circle in your graph "
    ).value(False)

    hide_system_pipelines = parameter(
        description="Hides the scheduled job launcher and driver submit pipelines at the API level to prevent clutter",
        default=True,
    )

    fix_env_on_osx = parameter(
        description="add no_proxy=* to env vars, fixing issues with multiprocessing on osx"
    )[bool]

    plugins = parameter(
        description="plugins to load on databand context creation", default=None
    )[str]
    allow_vendored_package = parameter(
        description="Allow adding dbnd/_vendor_package to sys.path", default=False
    )[bool]

    def _validate(self):
        if not self.databand_url and self.tracker_url:
            logger.warning(
                "core.databand_url was not set, using deprecated 'core.tracker_url' instead."
            )
            self.databand_url = self.tracker_url

        if self.databand_url and self.databand_url.endswith("/"):
            dbnd_log_debug(
                "Please fix your core.databand_url value, "
                "it should not contain '/' at the end, auto-fix has been applied."
            )
            self.databand_url = self.databand_url[:-1]

        # automatically disabling tracking if databand_url is not set
        if not self.databand_url:
            dbnd_log_debug(
                "Automatically disabling tracking to databand service as databand_url is not set"
            )
            self.tracker = [t for t in self.tracker if t != "api"]

        if self.databand_access_token and (self.dbnd_user or self.dbnd_password):
            logger.warning(
                "core.databand_access_token is used instead of defined dbnd_user and dbnd_password."
            )

    def build_tracking_store(self, remove_failed_store=True):
        from dbnd._core.tracking.registry import get_tracking_store

        return get_tracking_store(
            tracking_store_names=self.tracker,
            api_channel_name=self.tracker_api,
            tracker_raise_on_error=self.tracker_raise_on_error,
            remove_failed_store=remove_failed_store,
        )

    def build_databand_api_client(self):
        from dbnd.utils.api_client import ApiClient

        credentials = (
            {"token": self.databand_access_token}
            if self.databand_access_token
            else {"username": self.dbnd_user, "password": self.dbnd_password}
        )

        return ApiClient(self.databand_url, credentials=credentials)


class DynamicTaskConfig(Config):
    """
    Configuration section for dynamically generated tasks.
    """

    _conf__task_family = "dynamic_task"

    enabled = parameter(default=True, description="Allow tasks calls at runtime")[bool]
    in_memory_outputs = parameter(
        default=False, description="Store outputs for inline tasks in memory"
    )[bool]


class FeaturesConfig(Config):
    """
    Configuration section for caching features.
    """

    _conf__task_family = "features"

    in_memory_cache_target_value = parameter(
        default=True, description="Cache targets values in memory during execution"
    )[bool]


class TrackingConfig(Config):
    _conf__task_family = "tracking"

    databand_external_url = parameter(
        default=None,
        description="Tracker URL to be used for tracking from external systems",
    )[str]

    log_value_size = parameter(
        default=True,
        description="Calculate and log value size (can cause a full scan on not-indexable distributed memory objects) ",
    )[bool]

    log_value_schema = parameter(
        default=True, description="Calculate and log value schema "
    )[bool]
    log_value_stats = parameter(
        default=False,
        description="Calculate and log value stats(expensive to calculate, better use log_stats on parameter level)",
    )[bool]
    log_value_preview = parameter(
        default=True, description="Calculate and log value preview "
    )[bool]

    log_value_preview_max_len = parameter(
        description="Max size of value preview to be saved at DB, max value=50000"
    ).value(_DEFAULT_VALUE_PREVIEW_MAX_LEN)

    log_value_meta = parameter(
        default=True, description="Calculate and log value meta "
    )[bool]

    log_histograms = parameter(
        default=False,
        description="Enable calculation and tracking of histograms. Can be expensive",
    )[bool]

    auto_disable_slow_size = parameter(
        default=True,
        description="Auto disable slow preview for Spark DF with text formats",
    )[bool]

    flatten_operator_fields = parameter(
        default={},
        description="Control which of the operator's fields would be flatten when tracked",
    )[Dict[str, str]]

    def get_value_meta_conf(
        self, parameter_value_meta_conf, value_type=None, target=None
    ):
        # type: (ValueMetaConf, ValueType, Target) -> ValueMetaConf
        mc = parameter_value_meta_conf

        log_value_size = self.log_value_size
        if target and self.auto_disable_slow_size and log_value_size:
            log_value_size = value_type.support_fast_count(target)

        return ValueMetaConf(
            log_preview=first_not_none(mc.log_preview, self.log_value_preview),
            log_preview_size=first_not_none(
                mc.log_preview_size, self.log_value_preview_max_len
            ),
            log_schema=first_not_none(mc.log_schema, self.log_value_schema),
            log_size=first_not_none(mc.log_size, log_value_size),
            log_stats=first_not_none(mc.log_stats, self.log_value_stats),
            log_histograms=first_not_none(mc.log_histograms, self.log_histograms),
        )



from dbnd._core.cli.click_utils import _help
from dbnd._core.cli.service_auto_completer import completer
from dbnd._core.task_build.task_registry import get_task_registry
from dbnd._vendor import click


@click.command()
@click.argument("search", default="", autocompletion=completer.config())
@click.option("--module", "-m", help="Used for dynamic loading of modules")
@click.pass_context
def show_configs(ctx, module, search):
    """Show and search configurations"""
    _list_tasks(ctx, module, search, is_config=True)


@click.command()
@click.argument("search", default="", autocompletion=completer.task())
@click.option("--module", "-m", help="Used for dynamic loading of modules")
@click.pass_context
def show_tasks(ctx, module, search):
    """Show and search tasks"""
    _list_tasks(ctx, module, search, is_config=False)


COMMON_PARAMS = {"task_version", "task_env", "task_target_date"}


def _list_tasks(ctx, module, search, is_config):
    from dbnd import Config
    from dbnd._core.parameter.parameter_definition import _ParameterKind
    from dbnd._core.context.databand_context import new_dbnd_context

    formatter = ctx.make_formatter()

    with new_dbnd_context(module=module):
        tasks = get_task_registry().list_dbnd_task_classes()

    for task_cls in tasks:
        td = task_cls.task_definition
        full_task_family = td.full_task_family
        task_family = td.task_family

        if not (task_family.startswith(search) or full_task_family.startswith(search)):
            continue

        if issubclass(task_cls, Config) != is_config:
            continue

        dl = []
        for param_name, param_obj in td.all_task_params.items():
            if param_obj.system or param_obj.kind == _ParameterKind.task_output:
                continue
            if not is_config and param_name in COMMON_PARAMS:
                continue
            param_help = _help(param_obj.description)
            dl.append((param_name, param_help))

        if dl:
            with formatter.section(
                "{task_family} ({full_task_family})".format(
                    full_task_family=full_task_family, task_family=task_family
                )
            ):
                formatter.write_dl(dl)

    click.echo(formatter.getvalue().rstrip("\n"))



import logging

from dbnd._core.errors.base import DatabandConnectionException
from dbnd._core.errors.errors_utils import log_exception
from dbnd._core.tracking.backends import TrackingStore, TrackingStoreThroughChannel


logger = logging.getLogger(__name__)


class CompositeTrackingStore(TrackingStore):
    def __init__(self, tracking_stores, raise_on_error=True, remove_failed_store=False):
        if not tracking_stores:
            logger.warning("You are running without any tracking store configured.")
        self._stores = tracking_stores
        self._raise_on_error = raise_on_error
        self._remove_failed_store = remove_failed_store

    def _invoke(self, name, kwargs):
        res = None
        failed_stores = []
        for store in self._stores:
            try:
                handler = getattr(store, name)
                handler_res = handler(**kwargs)
                if handler_res:
                    res = handler_res
            except DatabandConnectionException as ex:
                logger.error(
                    "Failed to store tracking information from %s at %s : %s"
                    % (name, str(store), str(ex))
                )
                if self._remove_failed_store:
                    failed_stores.append(store)
                if self._raise_on_error:
                    raise
            except Exception as ex:
                if self._remove_failed_store:
                    failed_stores.append(store)
                log_exception(
                    "Failed to store tracking information from %s at %s: %s"
                    % (name, str(store), str(ex)),
                    ex,
                    non_critical=True,
                )
        if failed_stores:
            for store in failed_stores:
                logger.warning(
                    "Removing store %s from stores list due to failure" % (str(store),)
                )
                self._stores.remove(store)
        return res

    # this is a function that used for disabling Tracking api on spark inline tasks.
    def disable_tracking_api(self):
        filtered_stores = []

        for store in self._stores:
            if isinstance(store, TrackingStoreThroughChannel):
                continue
            filtered_stores.append(store)
        self._stores = filtered_stores

    def init_scheduled_job(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_scheduled_job.__name__, kwargs)

    def init_run(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_run.__name__, kwargs)

    def init_run_from_args(self, **kwargs):
        return self._invoke(CompositeTrackingStore.init_run_from_args.__name__, kwargs)

    def set_run_state(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_run_state.__name__, kwargs)

    def set_task_reused(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_reused.__name__, kwargs)

    def set_task_run_state(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_run_state.__name__, kwargs)

    def set_task_run_states(self, **kwargs):
        return self._invoke(CompositeTrackingStore.set_task_run_states.__name__, kwargs)

    def set_unfinished_tasks_state(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.set_unfinished_tasks_state.__name__, kwargs
        )

    def save_task_run_log(self, **kwargs):
        return self._invoke(CompositeTrackingStore.save_task_run_log.__name__, kwargs)

    def save_external_links(self, **kwargs):
        return self._invoke(CompositeTrackingStore.save_external_links.__name__, kwargs)

    def log_target(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_target.__name__, kwargs)

    def log_targets(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_targets.__name__, kwargs)

    def log_histograms(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_histograms.__name__, kwargs)

    def log_metrics(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_metrics.__name__, kwargs)

    def log_artifact(self, **kwargs):
        return self._invoke(CompositeTrackingStore.log_artifact.__name__, kwargs)

    def close(self):
        pass

    def add_task_runs(self, **kwargs):
        return self._invoke(CompositeTrackingStore.add_task_runs.__name__, kwargs)

    def heartbeat(self, **kwargs):
        return self._invoke(CompositeTrackingStore.heartbeat.__name__, kwargs)

    def save_airflow_task_infos(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.save_airflow_task_infos.__name__, kwargs
        )

    def update_task_run_attempts(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.update_task_run_attempts.__name__, kwargs
        )

    def is_ready(self, **kwargs):
        return all(store.is_ready() for store in self._stores)

    def save_airflow_monitor_data(self, **kwargs):
        return self._invoke(
            CompositeTrackingStore.save_airflow_monitor_data.__name__, kwargs
        )



import os

from dbnd._core.configuration.environ_config import ENV_DBND_USER


def _default_getuser():
    return "unknown"


def _find_getuser_f():
    try:
        if ENV_DBND_USER in os.environ:
            return lambda: os.environ[ENV_DBND_USER]

        if os.name == "nt":
            if hasattr(os, "getlogin"):
                os.getlogin()
                return os.getlogin

            import getpass

            if not hasattr(getpass, "getuser"):
                return _default_getuser()
            # check before we return it
            getpass.getuser()

        import getpass

        return getpass.getuser
    except Exception:
        return _default_getuser


dbnd_getuser = _find_getuser_f()



from dbnd._core.parameter.parameter_builder import (
    PARAMETER_FACTORY,
    ParameterFactory,
    build_parameter,
)
from dbnd._core.parameter.parameters_mapper import ParametersMapper
from targets.values import get_types_registry, get_value_type_of_type


_PARAMS_MAPPER = ParametersMapper()  # type: ParametersMapper


def get_params_mapper():
    return _PARAMS_MAPPER


def register_custom_parameter(value_type, parameter):
    value_type = get_value_type_of_type(value_type, inline_value_type=True)

    _PARAMS_MAPPER.register_custom_parameter(value_type, parameter)
    get_types_registry().register_value_type(value_type)

    return parameter


def get_parameter_for_value_type(value_type):
    return _PARAMS_MAPPER.get_parameter(value_type)



import sys

from dbnd._core.plugin.dbnd_plugins import is_airflow_enabled


######
## SUPPORT DBND TASK AS AIRFLOW OPERATORS
def is_in_airflow_dag_build_context():
    if "airflow" not in sys.modules:
        return False

    if not is_airflow_enabled():
        return False

    from dbnd_airflow.functional.dbnd_functional_dag import (
        is_in_airflow_dag_build_context as airflow__is_in_airflow_dag_build_context,
    )

    return airflow__is_in_airflow_dag_build_context()


def build_task_at_airflow_dag_context(task_cls, call_args, call_kwargs):
    from dbnd_airflow.functional.dbnd_functional_dag import (
        build_task_at_airflow_dag_context,
    )

    return build_task_at_airflow_dag_context(task_cls, call_args, call_kwargs)



import logging
import sys
import time

from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.utils.timeout import wait_until
from dbnd._vendor import click
from dbnd._vendor.click import command


logger = logging.getLogger(__name__)


@click.group(help="Tracker commands to ensure synchronization against webserver")
def tracker():
    pass


@tracker.command(help="Wait for the webserver to initialize")
@click.option(
    "--timeout",
    "-t",
    type=int,
    default=120,
    help="Amount of seconds to wait for webserver until timing out",
)
def wait(timeout):
    with new_dbnd_context(name="new_context") as dbnd_ctx:
        logger.info("Waiting {} seconds for tracker to become ready:".format(timeout))
        is_ready = wait_until(dbnd_ctx.tracking_store.is_ready, timeout)
        if not is_ready:
            logger.error("Tracker is not ready after {} seconds.".format(timeout))
            sys.exit(1)
        logger.info("Tracker is ready.")



from collections import defaultdict

import six

from dbnd._vendor import namesgenerator
from dbnd._vendor.namesgenerator import get_random_name


EXPORTABLE_DICT = defaultdict(get_random_name)


def get_exportable_value(value):
    return EXPORTABLE_DICT[str(value)]


def get_hashed_name(s, paranoid=False):
    import hashlib

    if paranoid:
        return get_exportable_value(s)

    digest = hashlib.md5(str(s).encode()).digest()
    if isinstance(digest, six.string_types):
        # python 2
        import struct

        digest = struct.unpack("16B", digest)

    a = digest[2] + 256 * digest[3]
    b = digest[0] + 256 * digest[1]
    name = "{}_{}".format(
        namesgenerator.left[a % len(namesgenerator.left)],
        namesgenerator.right[b % len(namesgenerator.right)],
    )

    return name



from typing import Tuple


_TERMINAL_SIZE = None


def get_terminal_size():
    # type: ()->Tuple[int,int]
    global _TERMINAL_SIZE
    if _TERMINAL_SIZE:
        return _TERMINAL_SIZE
    try:
        from click import get_terminal_size as click_terminal_size

        _TERMINAL_SIZE = click_terminal_size()
    except Exception:
        _TERMINAL_SIZE = (140, 25)
    return _TERMINAL_SIZE



import logging

import six

from more_itertools import unique_everseen
from six import iteritems

from dbnd._core.configuration.config_path import (
    CONF_CONFIG_SECTION,
    CONF_TASK_ENV_SECTION,
    CONF_TASK_SECTION,
)
from dbnd._core.configuration.config_value import ConfigValue
from dbnd._core.constants import ParamValidation, _TaskParamContainer
from dbnd._core.current import get_databand_context
from dbnd._core.decorator.task_decorator_spec import args_to_kwargs
from dbnd._core.errors import MissingParameterError, friendly_error
from dbnd._core.parameter import build_parameter, get_parameter_for_value_type
from dbnd._core.parameter.parameter_definition import (
    ParameterScope,
    build_parameter_value,
    infer_parameter_value_type,
)
from dbnd._core.parameter.parameter_value import ParameterValue
from dbnd._core.task_build.multi_section_config import MultiSectionConfig
from dbnd._core.task_build.task_context import try_get_current_task
from dbnd._core.task_build.task_params import TaskValueParams
from dbnd._core.task_build.task_passport import format_source_suffix
from dbnd._core.task_build.task_signature import TASK_ID_INVALID_CHAR_REGEX
from dbnd._core.task_ctrl.task_meta import TaskMeta
from dbnd._core.utils.basics.nothing import NOTHING
from dbnd._core.utils.basics.text_banner import safe_string
from targets import target
from targets.target_config import parse_target_config
from targets.values.builtins_values import DefaultObjectValueType


TASK_BAND_PARAMETER_NAME = "task_band"
logger = logging.getLogger(__name__)


class TaskFactoryConfig(object):
    """(Advanced) Databand's core task builder"""

    _conf__task_family = "task_build"

    def __init__(self, verbose, sign_with_full_qualified_name, sign_with_task_code):
        self.verbose = verbose
        self.sign_with_full_qualified_name = sign_with_full_qualified_name
        self.sign_with_task_code = sign_with_task_code

    @classmethod
    def from_dbnd_config(cls, conf):
        def _b(param_name):
            return conf.getboolean(cls._conf__task_family, param_name)

        return cls(
            verbose=_b("verbose"),
            sign_with_full_qualified_name=_b("sign_with_full_qualified_name"),
            sign_with_task_code=_b("sign_with_task_code"),
        )


def get_task_from_sections(config, task_name):
    extra_sections = []
    while task_name:
        # "pseudo" recursive call
        # we check if we have something  for current task_name, and if we do - that's a from
        task_from = config.get(task_name, "_from", None)
        if task_from == task_name:
            # let not throw exception, as it's not a critical error.
            task_from = None

        if task_from:
            extra_sections.append(task_from)
        task_name = task_from

    return extra_sections


class BaseTaskMetaFactory(object):
    def __init__(self, config, task_cls, task_args, task_kwargs):
        # type:(DbndConfig, Type[_BaseTask], Any, Any)->None
        self.task_cls = task_cls
        self.task_definition = task_cls.task_definition  # type: TaskDefinition

        # keep copy of user inputs
        self.task_kwargs__ctor = task_kwargs.copy()
        self.task_args__ctor = list(task_args)

        self.parent_task = try_get_current_task()

        # let find if we are running this constructor withing another Databand Task
        self.dbnd_context = get_databand_context()
        self.task_call_source = [
            self.dbnd_context.user_code_detector.find_user_side_frame(2)
        ]
        if self.task_call_source and self.parent_task:
            self.task_call_source.extend(self.parent_task.task_meta.task_call_source)

        self.task_family = self.task_definition.task_family
        self.task_name = self.task_family

        self.multi_sec_conf = MultiSectionConfig(config, [])

        self._task_params = self.task_definition._task_params.clone()

        self.ctor_kwargs = {}

        self._exc_desc = self.task_family
        self.task_errors = []

    def _build_task_params_values(self, task_params):
        param_task_env = None
        param_task_config = None
        param_def_regular = []
        # we want to build task_env and task_config parameters value first
        # so all others can use it
        for p in task_params.values():
            if p.name == "task_env":
                param_task_env = p
            elif p.name == "task_config":
                param_task_config = p
            else:
                param_def_regular.append(p)
        task_param_values = []
        if param_task_config:
            param_task_config_value = self.build_parameter_value(param_task_config)
            if param_task_config_value.value:
                # merging `Task.task_config` into current configuration
                self.multi_sec_conf.set_values(
                    param_task_config_value.value,
                    source=self._source_name("task_config"),
                )
            task_param_values.append(param_task_config_value)
        # if we have .task_env in the class
        # we need to calculate it first, so we can add it values to "root" configs
        if param_task_env:
            value_task_env = self.build_parameter_value(param_task_env)
            self._apply_task_env_config(task_env=value_task_env.value)
            task_param_values.append(value_task_env)
        # calculate configuration per parameter, and calculate parameter value
        regular_params = self.build_parameter_values(param_def_regular)
        task_param_values.extend(regular_params)
        return task_param_values

    def build_parameter_values(self, params):
        # type: (List[ParameterDefinition]) -> List[ParameterValue]
        result = []
        for param_def in params:
            try:
                p_value = self.build_parameter_value(param_def)
                result.append(p_value)
            except MissingParameterError as ex:
                self.task_errors.append(ex)
        return result

    def build_parameter_value(self, param_def):
        # This is the place we calculate param_def value
        # based on Class(defaults, constructor, overrides, root)
        # and Config(env, cmd line, config)  state

        # used for target_format update
        # change param_def definition based on config state
        param_def = self._update_param_def_target_config(param_def=param_def)
        param_name = param_def.name
        p_config_value = self.multi_sec_conf.get_param_config_value(param_def)

        if p_config_value and p_config_value.override:
            cf_value = p_config_value

        elif param_name in self.ctor_kwargs:
            cf_value = ConfigValue(
                self.ctor_kwargs.get(param_name), source=self._source_name("ctor")
            )
        elif p_config_value:
            cf_value = p_config_value
        elif param_def.is_output():
            # outputs can be none, we "generate" their values later
            cf_value = None
        else:
            err_msg = "No value defined for '{name}' at {context_str}".format(
                name=param_name, context_str=self._exc_desc
            )
            raise MissingParameterError(
                err_msg,
                help_msg=param_def._get_help_message(
                    sections=self.multi_sec_conf.sections
                ),
            )
        return build_parameter_value(param_def, cf_value)

    def _update_param_def_target_config(self, param_def):
        """calculates parameter.target_config based on extra config at parameter__target value"""
        target_option = "%s__target" % param_def.name
        target_config = self.multi_sec_conf.get_config_value(key=target_option)
        if not target_config:
            return param_def

        try:
            target_config = parse_target_config(target_config.value)
        except Exception as ex:
            raise param_def.parameter_exception(
                "Calculate target config for %s : target_config='%s'"
                % (target_option, target_config.value),
                ex,
            )

        param_def = param_def.modify(target_config=target_config)
        return param_def

    def _apply_task_env_config(self, task_env):
        # type: (EnvConfig)-> None
        # find same parameters in the current task class and EnvConfig
        # and take the value of that params from environment
        # for example  spark_config in SparkTask  (defined by EnvConfig.spark_config)

        # we take values using names only
        param_values = []
        for param_def in task_env._params.get_params():
            #  we want only parameters of the right scope -- children
            if param_def.scope != ParameterScope.children:
                continue
            param_values.append(
                (param_def.name, task_env._params.get_value(param_def.name))
            )
        self.multi_sec_conf.update_section(
            CONF_TASK_ENV_SECTION,
            param_values=param_values,
            source=self._source_name("env[%s]" % task_env.task_name),
        )

    def _source_name(self, name):
        return self.task_definition.task_passport.format_source_name(name)

    # should refactored out
    def _get_task_multi_section_config(self, config, task_kwargs):
        # there is priority of task name over task family, as name is more specific
        sections = [self.task_name]
        # _from at config files
        sections.extend(get_task_from_sections(config, self.task_name))
        sections.extend([self.task_family, self.task_definition.full_task_family])
        kwargs_task_config_sections = task_kwargs.pop("task_config_sections", None)
        if kwargs_task_config_sections:
            sections.extend(kwargs_task_config_sections)
        # adding "default sections"  - LOWEST PRIORITY
        if issubclass(self.task_definition.task_class, _TaskParamContainer):
            sections += [CONF_TASK_SECTION]
        from dbnd._core.task.config import Config

        if issubclass(self.task_definition.task_class, Config):
            sections += [CONF_CONFIG_SECTION]
        sections = list(unique_everseen(filter(None, sections)))

        return MultiSectionConfig(config, sections)


def _build_param(name):
    param_f = get_parameter_for_value_type(DefaultObjectValueType())
    param = build_parameter(param_f)
    param.name = name
    return param


class TaskMetaFactory(BaseTaskMetaFactory):
    """
     we have current config at  dbnd_config
     every tasks checks:
     1. overrides
     2. regular
     for (1) and (2) we are going to check following sections:
     A. it's own section    (task_family)
     if task have some child properties - it can affect it by "task" section


     1. overrides,
     2. constructor
     3. task_config
     4. configuration

     How to override specific task sub configs
     task_config = {  "spark" : { "param" : "ss"  }
     task_config = { spark.jars = some_jars ,
                     kubernetes.gpu = some_gpu }

    """

    def __init__(self, config, task_cls, task_args, task_kwargs):
        # type:(DbndConfig, Type[_BaseTask], Any, Any)->None
        super(TaskMetaFactory, self).__init__(
            config=config,
            task_cls=task_cls,
            task_args=task_args,
            task_kwargs=task_kwargs,
        )

        self.task_factory_config = TaskFactoryConfig.from_dbnd_config(config)
        self.verbose_build = self.task_factory_config.verbose

        self.task_family = task_kwargs.pop("task_family", self.task_family)
        # extra params from constructor
        self.task_name = task_kwargs.pop("task_name", None)

        self.task_config_override = task_kwargs.pop("override", None) or {}
        self.task_kwargs = task_kwargs

        if self.task_name:
            self.task_name = TASK_ID_INVALID_CHAR_REGEX.sub("_", self.task_name)

        # user gives explicit name, or it full_task_family
        self.task_main_config_section = (
            self.task_name or self.task_definition.task_config_section
        )

        if self.task_name is None:
            self.task_name = self.task_family

        self.multi_sec_conf = self._get_task_multi_section_config(config, task_kwargs)

        self.ctor_kwargs = None
        # utilities section
        self.build_warnings = []
        self._exc_desc = "%s(%s)" % (
            self.task_family,
            ", ".join(
                (
                    "%s=%s" % (p, safe_string(repr(k), 300))
                    for p, k in iteritems(self.task_kwargs__ctor)
                )
            ),
        )
        self.task_errors = []

    def create_dbnd_task_meta(self):
        # create task meta
        self._log_build_step(
            "Resolving task params with %s" % self.multi_sec_conf.sections
        )

        # let apply all "override" values in Task(override={})
        if self.task_config_override:
            self.multi_sec_conf.set_values(
                source=self._source_name("override"),
                config_values=self.task_config_override,
                override=True,
            )

        if self.verbose_build:
            # and issubclass(
            #    self.task_definition.task_class, _TaskParamContainer
            # ):
            self._log_config()
        self.ctor_kwargs = self._build_task_ctor_kwargs(
            self.task_args__ctor, self.task_kwargs
        )
        try:
            task_meta = self.build_task_meta()
        except Exception:
            if not self.verbose_build:
                self._log_config(force_log=True)
            raise
        self._log_build_step("Task Meta with obj_id = %s" % str(task_meta.obj_key))
        return task_meta

    def build_task_meta(self):
        # build list of all possible values
        task_param_values = self._build_task_params_values(self._task_params.all_params)

        validate_no_extra_params = next(
            (
                pv.value
                for pv in task_param_values
                if pv.name == "validate_no_extra_params"
            ),
            True,
        )
        if (
            validate_no_extra_params
            and validate_no_extra_params != ParamValidation.disabled
        ):
            self.validate_no_extra_config_params(validate_no_extra_params)

        self._assert_no_task_build_error()
        self._log_task_build_warnings()

        task_enabled = True
        if self.parent_task:
            task_enabled = self.parent_task.ctrl.should_run()

        task_band = self._get_task_band_value(task_param_values)
        if task_band:
            # we are going to load task from band
            task_enabled = False
            task_param_values = self.load_task_params_from_task_band(
                task_band, task_param_values
            )

        # update [task] section with Scope.children params
        self._apply_task_children_scope(task_param_values=task_param_values)

        return TaskMeta(
            task_definition=self.task_definition,
            task_family=self.task_family,
            task_name=self.task_name,
            task_params=TaskValueParams(task_param_values),
            task_config_override=self.task_config_override,
            config_layer=self.multi_sec_conf.layer,
            task_enabled=task_enabled,
            build_warnings=self.build_warnings,
            dbnd_context=self.dbnd_context,
            task_sections=self.multi_sec_conf.sections,
            task_call_source=self.task_call_source,
        )

    def validate_no_extra_config_params(self, validation_setting):
        """
        check that the user did not set any config values that don't have a matching param definition (protects against typos)
        """
        # Must lower task parameter name to comply to case insensitivity of configuration
        task_param_names = [
            tp.name.lower() for tp in self._task_params.all_params.values()
        ]
        for _, section in self.multi_sec_conf.iter_sections():
            for key, value in section.items():
                if (
                    key not in task_param_names
                    and not value.source.endswith(
                        format_source_suffix(ParameterScope.children.value)
                    )
                    and key not in ["_type", "_from"]
                    and not key.endswith("__target")
                ):
                    exc = friendly_error.task_build.unknown_parameter_in_config(
                        task_name=self.task_name,
                        param_name=key,
                        source=value.source,
                        task_param_names=task_param_names,
                        config_type=self._get_task_or_config_string(),
                    )
                    if validation_setting == ParamValidation.warn:
                        self.build_warnings.append(exc)
                    elif validation_setting == ParamValidation.error:
                        self.task_errors.append(exc)

    def _apply_task_children_scope(self, task_param_values):
        # type: (List[ParameterValue])-> None
        # we take values using names only
        if self.task_cls._conf__no_child_params:
            return
        param_values = []
        for p_val in task_param_values:
            #  we want only parameters of the right scope -- children
            if p_val.parameter.scope == ParameterScope.children:
                param_values.append((p_val.parameter.name, p_val.value))
        self.multi_sec_conf.update_section(
            CONF_TASK_SECTION,
            param_values=param_values,
            source=self._source_name(ParameterScope.children.value),
        )

    def _build_task_ctor_kwargs(self, task_args, task_kwargs):
        param_names = {p.name for p in self._task_params.all_params.values()}

        args_orig, kwargs_orig = list(task_args), task_kwargs.copy()
        has_varargs = False
        has_varkwargs = False
        if self.task_cls._conf__decorator_spec is not None:
            # only in functions we can have args as we know exact "call" signature
            task_args, task_kwargs = args_to_kwargs(
                self.task_cls._conf__decorator_spec.args, task_args, task_kwargs
            )
            has_varargs = self.task_cls._conf__decorator_spec.varargs
            has_varkwargs = self.task_cls._conf__decorator_spec.varkw

        # now we should not have any args, we don't know how to assign them
        if task_args and not has_varargs:
            raise friendly_error.unknown_args_in_task_call(
                self.parent_task, self.task_cls, func_params=(args_orig, kwargs_orig)
            )

        if not has_varkwargs:
            for param_name, _ in iteritems(task_kwargs):
                if param_name not in param_names:
                    raise friendly_error.task_build.unknown_parameter_in_constructor(
                        constructor=self._exc_desc,
                        param_name=param_name,
                        task_parent=self.parent_task,
                    )
        return task_kwargs

    def __str__(self):
        if self.task_name == self.task_family:
            return "TaskFactory(%s)" % self.task_name
        return "TaskFactory(%s@%s)" % (self.task_name, self.task_family)

    def _get_task_band_value(self, task_params):
        """
        returns task_band value ( means user has provided it to the class!)
        :param task_meta:
        :return:
        """
        for p_value in task_params:
            if p_value.name == TASK_BAND_PARAMETER_NAME and p_value.value:
                return p_value.value
        return None

    def load_task_params_from_task_band(self, task_band, task_params):
        task_band_value = target(task_band).as_object.read_json()

        new_params = []
        found = []

        source = "task_band.json"
        for p_value in task_params:
            if p_value.name not in task_band_value or p_value.name == "result":
                new_params.append(p_value)
                continue

            value = p_value.parameter.calc_init_value(task_band_value[p_value.name])
            found.append(p_value.name)
            new_parameter_value = ParameterValue(
                parameter=p_value.parameter,
                source=source,
                source_value=value,
                value=value,
            )
            new_params.append(new_parameter_value)

        logger.info(
            "Loading task '{task_family}' from {task_band}:\n"
            "\tfields taken:\t{found}".format(
                task_family=self.task_family, task_band=task_band, found=",".join(found)
            )
        )
        return new_params

    def _log_build_step(self, msg, force_log=False):
        if self.verbose_build or force_log:
            logger.info("[%s] %s", self.task_name, msg)

    def _log_config(self, force_log=False):
        msg = self.multi_sec_conf.config_log
        self._log_build_step(msg, force_log=force_log)

    def _assert_no_task_build_error(self):
        if not self.task_errors:
            return
        if len(self.task_errors) == 1:
            raise self.task_errors[0]
        raise friendly_error.failed_to_create_task(
            self._exc_desc, nested_exceptions=self.task_errors
        )

    def _get_task_or_config_string(self):
        from dbnd._core.task.config import Config

        if issubclass(self.task_definition.task_class, Config):
            return "config"
        else:
            return "task"

    def _log_task_build_warnings(self):
        if not self.build_warnings:
            return
        w = "Build warnings for %s '%s': " % (
            self._get_task_or_config_string(),
            self.task_name,
        )
        for warning in self.build_warnings:
            w += "\n\t" + str(warning)
            if hasattr(warning, "did_you_mean") and warning.did_you_mean:
                w += "\n\t\t - " + warning.did_you_mean
        logger.warning(w)


class TrackedTaskMetaFactory(BaseTaskMetaFactory):
    def __init__(self, config, task_cls, task_args, task_kwargs):
        # type:(DbndConfig, Type[_BaseTask], Any, Any)->None
        super(TrackedTaskMetaFactory, self).__init__(
            config=config,
            task_cls=task_cls,
            task_args=task_args,
            task_kwargs=task_kwargs,
        )

        self.multi_sec_conf = self._get_task_multi_section_config(config, task_kwargs)

    def create_dbnd_task_meta(self):
        task_args, task_kwargs = self.task_args__ctor, self.task_kwargs__ctor
        if self.task_cls._conf__decorator_spec is not None:
            task_args, task_kwargs = args_to_kwargs(
                self.task_cls._conf__decorator_spec.args, task_args, task_kwargs
            )

        # build list of all possible values
        class_task_param_values = self._build_task_params_values(
            self._task_params.class_params
        )
        user_param_values, user_results = self._build_user_parameter_values(
            task_args, task_kwargs
        )

        return TaskMeta(
            task_definition=self.task_definition,
            task_family=self.task_family,
            task_name=self.task_name,
            task_params=TaskValueParams(
                class_params=class_task_param_values,
                user_params=user_param_values,
                user_result_params=user_results,
            ),
            task_config_override={},
            config_layer=self.multi_sec_conf.layer,
            dbnd_context=self.dbnd_context,
            task_sections=self.multi_sec_conf.sections,
            task_call_source=self.task_call_source,
        )

    def _build_user_parameter_values(self, task_args, task_kwargs):
        # TODO: should we add somehow anything from task_args, rest of task_kwargs?
        values = {}
        for i, value in enumerate(task_args):
            values[str(i)] = self._build_user_parameter_value(
                _build_param(str(i)), value
            )

        for param_def in self._task_params.user_params.values():
            if param_def.name in task_kwargs:
                value = task_kwargs.get(param_def.name, NOTHING)
                values[param_def.name] = self._build_user_parameter_value(
                    param_def, value
                )

        for name, value in six.iteritems(task_kwargs):
            if name not in values:
                values[name] = self._build_user_parameter_value(
                    _build_param(name), value
                )

        user_results = []
        for param_def in self._task_params._user_result_params.values():
            user_results.append(self._build_user_parameter_value(param_def, NOTHING))

        return list(values.values()), user_results

    def _build_user_parameter_value(self, param_def, value):
        if value is NOTHING:
            parameter, warnings = param_def, []
            actual_value = param_def.default
        else:
            parameter, warnings = infer_parameter_value_type(param_def, value)
            actual_value = value

        return ParameterValue(
            parameter=parameter,
            source=self.task_definition.full_task_family_short,
            source_value=value,
            value=actual_value,
            parsed=False,
            warnings=warnings,
        )



import logging

from time import sleep

import requests
import six

from six.moves.urllib_parse import urljoin

from dbnd._core.errors.base import DatabandApiError, DatabandConnectionException
from dbnd._core.errors.friendly_error.api import api_connection_refused
from dbnd._core.utils.http.retry_policy import LinearRetryPolicy


logger = logging.getLogger(__name__)


# uncomment for requests trace
# import http.client
# http.client.HTTPConnection.debuglevel = 1


class ApiClient(object):
    """Json API client implementation."""

    api_prefix = "/api/v1/"
    default_headers = None

    def __init__(self, api_base_url, credentials=None):
        self._api_base_url = api_base_url
        self.is_auth_required = bool(credentials)
        self.credentials = credentials
        self.session = None
        self.default_headers = {"Accept": "application/json"}

    def _request(self, endpoint, method="GET", data=None, headers=None, query=None):
        if not self.session:
            self._init_session(self.credentials)

        headers = dict(self.default_headers, **(headers or {}))

        url = urljoin(self._api_base_url, endpoint)
        try:
            resp = self.session.request(
                method=method, url=url, json=data, headers=headers, params=query
            )
        except requests.exceptions.ConnectionError:
            self.session = None
            raise

        if not resp.ok:
            raise DatabandApiError(
                method, url, resp.status_code, resp.content.decode("utf-8")
            )
        if resp.content:
            try:
                return resp.json()
            except Exception:
                return None
        return resp.json() if resp.content else None

    def _init_session(self, credentials):
        try:
            self.session = requests.session()

            if not self.is_auth_required:
                return

            token = credentials.get("token")
            if token:
                self.default_headers["Authorization"] = "Bearer {}".format(token)
                return

            # get the csrf token cookie (if enabled on the server)
            self.session.get(urljoin(self._api_base_url, "/app"))
            csrf_token = self.session.cookies.get("dbnd_csrftoken")
            if csrf_token:
                self.default_headers["X-CSRFToken"] = csrf_token

            if "username" in credentials and "password" in credentials:
                self.api_request(
                    "auth/login", method="POST", data=credentials,
                )
            else:
                logger.warning(
                    "ApiClient._init_session: username or password is not provided"
                )
        except Exception:
            self.session = None
            raise

    def api_request(
        self,
        endpoint,
        data,
        method="POST",
        headers=None,
        query=None,
        no_prefix=False,
        retry_policy=None,
        failure_handler=None,
    ):
        retry_number = 0
        retry_policy = retry_policy or LinearRetryPolicy(0, 1)  # single chance
        url = endpoint if no_prefix else urljoin(self.api_prefix, endpoint)
        while True:
            retry_number += 1
            try:
                resp = self._request(
                    url, method=method, data=data, headers=headers, query=query
                )
            except requests.ConnectionError as ex:
                if failure_handler:
                    failure_handler(ex, retry_policy, retry_number)
                if retry_policy.should_retry(500, None, retry_number):
                    sleep(retry_policy.seconds_to_sleep(retry_number))
                    continue
                raise api_connection_refused(self._api_base_url + url, ex)
            return resp

    def is_ready(self):
        try:
            is_auth_required = self.is_auth_required
            self.is_auth_required = False
            self.api_request("auth/ping", None, method="GET")
            return True
        except (DatabandConnectionException, DatabandApiError):
            return False
        finally:
            self.is_auth_required = is_auth_required


def dict_dump(obj_dict, value_schema):
    """
    Workaround around marshmallow 2.0 not supporting Dict with types
    :return:
    """
    return {
        key: value_schema.dump(value).data for key, value in six.iteritems(obj_dict)
    }



# It differs from dbnd._vendor in a way _vendor_package package is used:
# instead of replacing all imports to dbnd._vendor, we add path to _vendor_package
# module to `sys.path`. This approach might casue versions
# conflicts that we agreed to solve one by one.
#



import inspect

import attr
import six

from dbnd._core.utils.basics.nothing import NOTHING, is_defined
from dbnd._core.utils.typing_utils.doc_annotations import get_doc_annotaions
from dbnd._vendor.pytypes.type_utils import get_Tuple_params, is_Tuple
from targets.values import get_types_registry


@attr.s
class _TaskDecoratorSpec(object):
    item = attr.ib()
    is_class = attr.ib()

    decorator_kwargs = attr.ib()
    default_result = attr.ib()

    # func spec
    args = attr.ib()
    varargs = attr.ib()
    varkw = attr.ib()
    defaults = attr.ib()
    kwonlyargs = attr.ib()
    kwonlydefaults = attr.ib()
    annotations = attr.ib()
    defaults_values = attr.ib()
    name = attr.ib()
    doc_annotations = attr.ib()


# Extracts arg types and return type from doc string
# # type: (str, dict) -> int
#          ^^^^^^^^^     ^^^

# Will split args type string to types (one level of nested commas supported)
# "str, dict(str, bool), list[str]" => ["str", "dict(str, bool)", "list[str]"]


def build_task_decorator_spec(class_or_func, decorator_kwargs, default_result):
    # type: (callable, ... ) -> _TaskDecoratorSpec
    is_class = inspect.isclass(class_or_func)
    if is_class:
        f = class_or_func.__init__
    else:
        f = class_or_func

    name = class_or_func.__name__
    if six.PY3:
        # python 3 https://docs.python.org/3/library/inspect.html#inspect.getfullargspec
        f_spec = inspect.getfullargspec(f)
        varkw = f_spec.varkw
        kwonlyargs = f_spec.kwonlyargs
        kwonlydefaults = f_spec.kwonlydefaults
        annotations = f_spec.annotations
    else:
        # python 2
        f_spec = inspect.getargspec(f)
        varkw = f_spec.keywords
        kwonlyargs = None
        kwonlydefaults = None
        annotations = None

    f_defaults = f_spec.defaults or {}
    defaults_start_at = len(f_spec.args) - len(f_defaults)

    return _TaskDecoratorSpec(
        item=class_or_func,
        is_class=is_class,
        decorator_kwargs=decorator_kwargs,
        default_result=default_result,
        args=f_spec.args,
        varargs=f_spec.varargs,
        varkw=varkw,
        defaults={
            f_spec.args[defaults_start_at + i]: v for i, v in enumerate(f_defaults)
        },
        kwonlyargs=kwonlyargs or [],
        kwonlydefaults=kwonlydefaults or {},
        annotations=annotations or {},
        defaults_values=f_defaults,
        name=name,
        doc_annotations=get_doc_annotaions(f, f_spec.args) or {},
    )


# 1. read default values provided and get their type - type(default_value), where exist
# 2. read py3 annotations if exist, and get args types
# 3. read doc string and get args types as str (doc string not necessary will compile)
# 4. for each one of the above - guess which Parameter type it should be. Currently supported:
#  a. bool/int/float/str -> all convert to relevant *Parameter type
#  b. DataFrame will convert to DataFrameParameter
#  c. collection types (currently only list, dict, tuple, set) won't convert to anything
#     (theoretically, it could be or *Param or TargetParameter)
#  d. if non of the above matched convert to Parameter
#  TODO: this logic should be certainly improved, even maybe give option to extend those conversions (someday)
# 5. If only single type "guessed" - use it. If more than one type (i.e doc str returned as_pickle,
#    but default_value returned float) - nothing happens, probably should raise exception. If 0 types
#    guessed - assume simple Parameter
def guess_func_arg_value_type(f_spec, name, default_value):
    # type: (_TaskDecoratorSpec, str, Any) -> ValueType
    r = get_types_registry()

    annotation = f_spec.annotations.get(name)
    if annotation is not None:
        return r.get_value_type_of_type(annotation)

    doc_annotation = f_spec.doc_annotations.get(name)
    t = r.get_value_type_of_type_str(doc_annotation)
    if t:
        return t

    if is_defined(default_value):
        return r.get_value_type_of_type(type(default_value))

    return None


def guess_func_return_type(f_spec):
    return_key = "return"

    r = get_types_registry()

    return_func_spec = f_spec.annotations.get(return_key, NOTHING)
    if is_defined(return_func_spec):
        if return_func_spec is None:
            return None

        # for -> (int,DataFrame)
        if isinstance(return_func_spec, tuple):
            return_func_spec = [
                ("result_%s" % idx, ret_part_type)
                for idx, ret_part_type in enumerate(return_func_spec, start=1)
            ]
        # easy way to check that it's NamedTuple
        elif hasattr(return_func_spec, "_field_types"):
            return_func_spec = list(six.iteritems(return_func_spec._field_types))
        # case of named tuple
        elif is_Tuple(return_func_spec):
            # for -> Tuple[int,DataFrame]
            return_func_spec = get_Tuple_params(return_func_spec)
            return_func_spec = [
                ("result_%s" % idx, ret_part_type)
                for idx, ret_part_type in enumerate(return_func_spec, start=1)
            ]

        if isinstance(return_func_spec, list):
            result = []
            for field_name, ret_part_type in return_func_spec:
                field_value_type = r.get_value_type_of_type(
                    ret_part_type, inline_value_type=True
                )
                result.append((field_name, field_value_type))
            return result
        else:
            # fallback to regular parsing
            return r.get_value_type_of_type(return_func_spec, inline_value_type=True)

    doc_annotation = f_spec.doc_annotations.get(return_key, NOTHING)
    if is_defined(doc_annotation):
        if doc_annotation == "None":
            return None
        # if it "return" , it parsed into list
        # doc_annotation
        if isinstance(doc_annotation, six.string_types):
            return r.get_value_type_of_type_str(doc_annotation) or NOTHING

        # so we have multiple params
        result = []
        for idx, ret_part_type in enumerate(doc_annotation, start=1):
            field_value_type = r.get_value_type_of_type_str(ret_part_type)
            result.append(("result_%s" % idx, field_value_type))
        return result
    return NOTHING


def args_to_kwargs(arg_names, args, kwargs):
    if not args:
        return args, kwargs

    # convert args to kwargs
    zipped = dict(zip(arg_names, args))
    # only if none of zipped keys passed in kwargs
    if all(k not in kwargs for k in zipped.keys()):
        n = len(zipped)
        args = args[n:]
        kwargs = kwargs.copy()
        kwargs.update(zipped)

    return args, kwargs



import logging
import os

from dbnd._core.configuration.environ_config import reset_dbnd_project_config


logger = logging.getLogger(__name__)


TEST_DBND_HOME = "/tmp/test-dbnd"


def pytest_configure(config):
    if not os.path.exists(TEST_DBND_HOME):
        os.mkdir(TEST_DBND_HOME)
    os.environ["DBND_HOME"] = TEST_DBND_HOME
    reset_dbnd_project_config()



"""
All code related to "getting" file from project based on databand library
Deprecated in favor for dbnd_project_config at dbnd._core.configuration.environ_config
"""

from dbnd._core.configuration.environ_config import get_dbnd_project_config
from dbnd._core.utils.basics.path_utils import (
    abs_join,
    relative_path,
    relative_path_directory,
)


def databand_lib_path(*path):
    return get_dbnd_project_config().dbnd_lib_path(*path)


def databand_config_path(*path):
    return get_dbnd_project_config().dbnd_config_path(*path)


def databand_system_path(*path):
    return get_dbnd_project_config().dbnd_system_path(*path)


def project_path(*path):
    return get_dbnd_project_config().dbnd_project_path(*path)


__all__ = [
    "abs_join",
    "relative_path_directory",
    "relative_path",
    "databand_lib_path",
    "databand_config_path",
    "databand_system_path",
    "project_path",
]



import atexit
import logging
import os
import sys
import typing

from subprocess import list2cmdline
from typing import Optional

from dbnd._core.configuration import get_dbnd_project_config
from dbnd._core.configuration.dbnd_config import config
from dbnd._core.constants import RunState, TaskRunState, UpdateSource
from dbnd._core.context.databand_context import new_dbnd_context
from dbnd._core.current import is_verbose, try_get_databand_run
from dbnd._core.errors.errors_utils import UserCodeDetector
from dbnd._core.inplace_run.airflow_dag_inplace_tracking import (
    build_run_time_airflow_task,
    override_airflow_log_system_for_tracking,
    try_get_airflow_context,
)
from dbnd._core.run.databand_run import new_databand_run
from dbnd._core.task.task import Task
from dbnd._core.task_run.task_run import TaskRun
from dbnd._core.task_run.task_run_error import TaskRunError
from dbnd._core.utils.timezone import utcnow
from dbnd._vendor import pendulum


logger = logging.getLogger(__name__)

if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.run.databand_run import DatabandRun

    T = typing.TypeVar("T")


def set_tracking_config_overide(airflow_context=None, use_dbnd_log=None):
    # 1. create proper DatabandContext so we can create other objects
    track_with_cache = config.getboolean("run", "tracking_with_cache")
    config_for_tracking = {
        "run": {
            "skip_completed": track_with_cache,
            "skip_completed_on_run": track_with_cache,
            "validate_task_inputs": track_with_cache,
            "validate_task_outputs": track_with_cache,
        },  # we don't want to "check" as script is task_version="now"
        "task": {
            "task_in_memory_outputs": not track_with_cache
        },  # do not save any outputs
        "core": {"tracker_raise_on_error": False},  # do not fail on tracker errors
    }
    if airflow_context:
        import pytz

        task_target_date = pendulum.parse(
            airflow_context.execution_date, tz=pytz.UTC
        ).date()
        use_dbnd_log = override_airflow_log_system_for_tracking()
        config_for_tracking["task"]["task_target_date"] = task_target_date

    if use_dbnd_log is not None:
        config_for_tracking["log"] = {"disabled": not use_dbnd_log}
    return config.set_values(
        config_values=config_for_tracking, override=True, source="dbnd_tracking_config"
    )


class _DbndInplaceRunManager(object):
    def __init__(self):
        self._context_managers = []
        self._atexit_registered = False

        self._active = False

        self._run = None
        self._task_run = None

    def _enter_cm(self, cm):
        # type: (typing.ContextManager[T]) -> T
        # else contextManagers are getting closed sometimes :(
        val = cm.__enter__()
        self._context_managers.append(cm)
        return val

    def _close_all_context_managers(self):
        while self._context_managers:
            cm = self._context_managers.pop()
            cm.__exit__(None, None, None)

    def start(self, root_task_name=None, airflow_context=None):
        if self._run or self._active or try_get_databand_run():
            return

        airflow_context = airflow_context or try_get_airflow_context()
        set_tracking_config_overide(use_dbnd_log=True, airflow_context=airflow_context)

        dc = self._enter_cm(new_dbnd_context())  # type: DatabandContext

        if airflow_context:
            root_task, job_name, source = build_run_time_airflow_task(airflow_context)
        else:
            root_task = _build_inline_root_task(root_task_name)
            job_name = None
            source = UpdateSource.dbnd

        self._run = self._enter_cm(
            new_databand_run(
                context=dc,
                task_or_task_name=root_task,
                job_name=job_name,
                existing_run=False,
                source=source,
                af_context=airflow_context,
                send_heartbeat=False,
            )
        )  # type: DatabandRun

        if not self._atexit_registered:
            _set_process_exit_handler(self.stop)
            self._atexit_registered = True

        sys.excepthook = self.stop_on_exception
        self._active = True

        # now we send data to DB
        self._run._init_without_run()
        self._start_taskrun(self._run.driver_task_run)
        self._start_taskrun(self._run.root_task_run)
        self._task_run = self._run.root_task_run

        return self._task_run

    def _start_taskrun(self, task_run):
        self._enter_cm(task_run.runner.task_run_execution_context())
        task_run.set_task_run_state(state=TaskRunState.RUNNING)

    def stop(self):
        if not self._active:
            return
        self._active = False
        try:
            databand_run = self._run
            root_tr = self._task_run
            root_tr.finished_time = utcnow()

            if root_tr.task_run_state not in TaskRunState.finished_states():
                for tr in databand_run.task_runs:
                    if tr.task_run_state == TaskRunState.FAILED:
                        root_tr.set_task_run_state(TaskRunState.UPSTREAM_FAILED)
                        break
                else:
                    root_tr.set_task_run_state(TaskRunState.SUCCESS)

            driver_tr = databand_run.driver_task.current_task_run
            if driver_tr.task_run_state not in TaskRunState.finished_states():
                driver_tr.set_task_run_state(TaskRunState.SUCCESS)

            if root_tr.task_run_state == TaskRunState.SUCCESS:
                databand_run.set_run_state(RunState.SUCCESS)
            else:
                databand_run.set_run_state(RunState.FAILED)
            logger.info(databand_run.describe.run_banner_for_finished())

            self._close_all_context_managers()
        except Exception as ex:
            _handle_inline_run_error("dbnd-tracking-shutdown")

    def stop_on_exception(self, type, value, traceback):
        if self._active:
            try:
                error = TaskRunError.build_from_ex(
                    ex=value, task_run=self._task_run, exc_info=(type, value, traceback)
                )
                self._task_run.set_task_run_state(TaskRunState.FAILED, error=error)
            except:
                _handle_inline_run_error("dbnd-set-script-error")

        self.stop()
        sys.__excepthook__(type, value, traceback)


def _set_process_exit_handler(handler):
    atexit.register(handler)

    # https://docs.python.org/3/library/atexit.html
    # The functions registered via this module are not called when the program
    # is killed by a signal not handled by Python, when a Python fatal internal
    # error is detected, or when os._exit() is called.
    #                       ^^^^^^^^^^^^^^^^^^^^^^^^^
    # and os._exit is the one used by airflow (and maybe other libraries)
    # so we'd like to monkey-patch os._exit to stop dbnd inplace run manager
    original_os_exit = os._exit

    def _dbnd_os_exit(*args, **kwargs):
        try:
            handler()
        finally:
            original_os_exit(*args, **kwargs)

    os._exit = _dbnd_os_exit


def _build_inline_root_task(root_task_name):
    # create "root task" with default name as current process executable file name
    if not root_task_name:
        root_task_name = sys.argv[0].split(os.path.sep)[-1]

    class InplaceTask(Task):
        _conf__task_family = root_task_name

    try:
        user_frame = UserCodeDetector.build_code_detector().find_user_side_frame(
            user_side_only=True
        )
        if user_frame:
            module_code = open(user_frame.filename).read()
            InplaceTask.task_definition.task_module_code = module_code
            InplaceTask.task_definition.task_source_code = module_code
    except Exception as ex:
        logger.info("Failed to find source code: %s", str(ex))

    root_task = InplaceTask(task_version="now", task_name=root_task_name)

    root_task.task_meta.task_command_line = list2cmdline(sys.argv)
    root_task.task_meta.task_functional_call = "bash_cmd(args=%s)" % repr(sys.argv)

    return root_task


def try_get_inplace_task_run():
    # type: ()->Optional[TaskRun]
    if get_dbnd_project_config().is_tracking_mode():
        return dbnd_run_start()


# there can be only one tracking manager
_dbnd_inline_manager = None  # type: Optional[_DbndInplaceRunManager]


def dbnd_run_start(name=None, airflow_context=None):
    if get_dbnd_project_config().disabled:
        return None

    global _dbnd_inline_manager
    if not _dbnd_inline_manager:
        dsm = _DbndInplaceRunManager()
        try:
            dsm.start(name, airflow_context)

            if dsm._active:
                _dbnd_inline_manager = dsm
        except Exception as e:
            logger.error(e, exc_info=True)
            _handle_inline_run_error("inline-start")
            get_dbnd_project_config().disabled = True
            return None
    if _dbnd_inline_manager and _dbnd_inline_manager._active:
        return _dbnd_inline_manager._task_run


def dbnd_run_stop():
    global _dbnd_inline_manager
    if _dbnd_inline_manager:
        _dbnd_inline_manager.stop()
        _dbnd_inline_manager = None


def _handle_inline_run_error(msg):
    if is_verbose():
        logger.warning(
            "Failed during dbnd %s, ignoring, and continue without tracking" % msg,
            exc_info=True,
        )
    else:
        logger.info(
            "Failed during dbnd %s, ignoring, and continue without tracking" % msg
        )



import logging
import typing

from typing import Optional

from dbnd._core.configuration import get_dbnd_project_config


if typing.TYPE_CHECKING:
    from dbnd._core.context.databand_context import DatabandContext
    from dbnd._core.run.databand_run import DatabandRun
    from dbnd._core.task_run.task_run import TaskRun
    from dbnd._core.settings import DatabandSettings
    from dbnd._core.task import Task


def get_databand_context():
    # type: () -> DatabandContext

    from dbnd._core.context.databand_context import DatabandContext  # noqa: F811

    return DatabandContext.try_instance()


def try_get_databand_context():
    # type: () -> Optional[DatabandContext]

    from dbnd._core.context.databand_context import DatabandContext  # noqa: F811

    if not DatabandContext.has_instance():
        return None
    return get_databand_context()


def dbnd_context():
    context = try_get_databand_context()
    if not context:
        # we are running without Databand Context
        # let create one inplace
        from dbnd._core.context.databand_context import DatabandContext  # noqa: F811

        context = DatabandContext.try_instance(name="inplace_run")
    return context


def get_databand_run():
    # type: () -> DatabandRun
    from dbnd._core.run.databand_run import DatabandRun  # noqa: F811

    v = DatabandRun.get_instance()
    return v


def try_get_databand_run():
    # type: () -> Optional[DatabandRun]
    from dbnd._core.run.databand_run import DatabandRun  # noqa: F811

    if DatabandRun.has_instance():
        return get_databand_run()
    return None


def current_task():
    # type: () -> Task
    from dbnd._core.task_build.task_context import current_task as ct

    return ct()


def try_get_current_task():
    from dbnd._core.task_build.task_context import try_get_current_task as tgct

    return tgct()


def get_task_by_task_id(task_id):
    # type: (str) -> Task
    return get_databand_context().task_instance_cache.get_task_by_id(task_id)


def try_get_current_task_run():
    # type: () -> TaskRun
    run = try_get_databand_run()
    if not run:
        return None
    task = try_get_current_task()
    if task:
        return run.get_task_run(task.task_id)
    else:
        return None


def current_task_run():
    # type: () -> TaskRun
    return get_databand_run().get_task_run(current_task().task_id)


def get_settings():
    # type: () -> DatabandSettings
    from dbnd._core.settings import DatabandSettings  # noqa: F811

    v = get_databand_context().settings  # type: DatabandSettings
    return v


def is_verbose():
    context = try_get_databand_context()
    if context and getattr(context, "system_settings", None):
        if context.system_settings.verbose:
            # only if True, otherwise check project config too
            return True

    return get_dbnd_project_config().is_verbose()


def get_target_logging_level():
    default_level = 10  # DEBUG

    context = try_get_databand_context()
    if context and getattr(context, "settings", None):
        return getattr(logging, context.settings.log.targets_log_level)

    return default_level


def is_killed():
    run = try_get_databand_run()
    return run and run.is_killed()
